//------------------------------------------------------------------------------
// pong.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file: pong.cpp
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 09.09.2009
 */

#include <gewoopi/gewoopi.hpp>
#include <gewoopi/scene.hpp>
#include <gewoopi/input.hpp>

#include <gewoopi/scene/mesh_manager.hpp>
#include <gewoopi/video/renderer.h>
#include <gewoopi/video/material.hpp>

int main( int argc, char *argv[] )
{

  using namespace gewoopi;

  core::Application game("pong", argc, argv);

  // create a perspective camera
  scene::CameraPtr cam(
    new scene::CameraPerspective( math::Degreef(40),    // field of view
                                  4./3,                 // aspect ratio,
                                  0.7,                  // near clipping
                                  40 ) );               // far clipping
  game.getRenderer()->setCamera( cam );
  //game.getRenderer()->setCullMode( gewoopi::video::CCW );

  scene::MeshManager* mesh_loader = scene::MeshManager::instance();
  scene::ManagerPtr scene_manager = game.getSceneManager();

  scene::Tree::iterator paddle  = scene_manager->createNode();
  scene::Tree::iterator pong    = scene_manager->createNode();
  scene::Tree::iterator sphere  = scene_manager->createNode();

  paddle->addMesh(mesh_loader->load("data/paddle.obj"))->calculateCollisionData();
  pong->addMesh(  mesh_loader->load("data/pong.obj")  )->calculateCollisionData();
  sphere->addMesh(mesh_loader->load("data/sphere.obj"));

  //----------------------------------------------------------------------------

  input::Button quit(  game.getInputManager()->getButton("quit" ) );
  input::Button pause( game.getInputManager()->getButton("break") );

  input::Button left ( game.getInputManager()->getButton("left" ) );
  input::Button right( game.getInputManager()->getButton("right") );
  input::Button speed( game.getInputManager()->getButton("speed") );

  input::Button cam_forward(game.getInputManager()->getButton("cam-forward" ) );
  input::Button cam_back(   game.getInputManager()->getButton("cam-back"    ) );

  math::Vector3f paddle_pos;
  math::Vector3f sphere_pos(0,0.5,-1.5);
  math::Vector3f sphere_dir(0.8,0,-1);
  sphere_dir.normalize();
  math::Vector3f sphere_move;

  math::Vector3f cam_pos( math::Vector3f(0,5,3) );
  math::Vector3f cam_look_at( math::Vector3f(0,0,-5) );

  float paddle_speed = 0;

  while( game.run() && !quit )
  {

    if( !pause )
    {
      //--------------------------------------------------------------------------
      // paddle
      //

      paddle_speed += (right - left) * 20 * game.getLastFrameTime();
      paddle_speed *= std::pow(0.2, game.getLastFrameTime());

      if( std::fabs(paddle_speed) < 0.1 )
        paddle_speed = 0;

      if( paddle_speed > 20 )
        paddle_speed = 20;
      else if ( paddle_speed < -20 )
        paddle_speed = -20;

      paddle_pos.x() += paddle_speed
                      * (1 + speed)
                      * game.getLastFrameTime();

      if( paddle_pos.x() <= -9 )
      {
        paddle_pos.x() = -9;
        paddle_speed *= -0.5;
      }
      else if( paddle_pos.x() >= 9 )
      {
        paddle_pos.x() = 9;
        paddle_speed *= -0.5;
      }

      paddle->setPosition( paddle_pos );

      //------------------------------------------------------------------------
      // sphere
      //

      sphere_move = sphere_dir * 11 * game.getLastFrameTime();

      core::real32 dist;
      math::Vector3f check_dir;

      //-----------------
      // check collisions
      //-----------------

      if( sphere_dir.x() < 0 )
        check_dir = math::Vector3f(-1,0,0);
      else
        check_dir = math::Vector3f(1,0,0);

      if( pong->hit(math::Ray3f(sphere_pos, check_dir), &dist) )
        if( dist - 0.5f <= fabs(sphere_move.x()) )
        {
          sphere_dir.x()  *= -1;
          sphere_move.x() *= -1;

          if( sphere_move.x() >= 0 )
            sphere_move.x() -= dist - 0.5f;
          else
            sphere_move.x() += dist - 0.5f;
        }

      if( sphere_dir.z() < 0 )
        check_dir = math::Vector3f(0,0,-1);
      else
        check_dir = math::Vector3f(0,0,1);

      if( pong->hit(math::Ray3f(sphere_pos, check_dir), &dist) )
      {
        if( dist - 0.5f <= fabs(sphere_move.z()) )
        {
          sphere_dir.z()  *= -1;
          sphere_move.z() *= -1;

          if( sphere_move.z() >= 0 )
            sphere_move.z() -= dist - 0.5f;
          else
            sphere_move.z() += dist - 0.5f;
        }
      }

      if( paddle->hit(math::Ray3f(sphere_pos, math::Vector3f(0,0,1)), &dist) )
      {
        if( sphere_dir.z() > 0 && dist - 0.5 <= sphere_move.z() )
        {
          sphere_dir.z() *= -1;
          sphere_move.z() *= -1;
          sphere_dir.x() -= 0.04 * paddle_speed; // add some spine
          sphere_dir.normalize();
        }
      }

      if( paddle->hit(math::Ray3f(sphere_pos, sphere_dir), &dist) )
      {
        if( sphere_dir.z() > 0 && dist - 0.5f <= sphere_move.length() )
        {
          sphere_dir *= -1;
          sphere_move *= -1;
        }
      }

      sphere_pos += sphere_move;

      if( sphere_pos.z() > 1.5 )
      {
        sphere_pos.z() = -8;
      }

      sphere->setPosition( sphere_pos );
    }

    //--------------------------------------------------------------------------
    // camera
    //

    cam_pos.x( paddle_pos.x() * 0.6f );
    cam_pos.z() += (cam_forward - cam_back) * 5 * game.getLastFrameTime();
    cam_look_at.x( paddle_pos.x() * 0.5f );

    cam->setTransform(
      math::Matrix4x4f::LookAtLH( cam_pos,
                                  cam_look_at,
                                  math::Vector3f(0,1,0) ) );
  }

  return 0;
}
