//------------------------------------------------------------------------------
// application.h - this file is part of the gewoopi game engine
// (c) copyright 2008-2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#ifndef APPLICATION_H_
#define APPLICATION_H_

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>

#include <xml-sipa/document.h>

#include <gewoopi/video/fwd.h>
#include <gewoopi/input/fwd.h>
#include <gewoopi/scene/fwd.h>

#include <gewoopi/math/vector.hpp>
#include <gewoopi/utilities/timer.h>
#include <gewoopi/utilities/dynamic_average.h>

#include <gewoopi/core/option.hpp>
#include <gewoopi/core/vfs/virtual_file_system.hpp>

namespace gewoopi
{
namespace core
{

  using boost::program_options::options_description;
  using boost::filesystem::path;

  /// @brief A simple bool wrapper for the auto quit option

  struct AutoQuit: public Option<bool>
  {
      AutoQuit( bool auto_quit ):
        Option<bool>( auto_quit )
      {};
  };

  /// @brief An application

  class Application
  {

    public:

      //------------------------------------------------------------------------
      ///
      /// @param name The name of the application
      /// @param argc The argument count
      /// @param argv An array of arguments
      /// @param auto_quit Whether the program should be quit if an error in the
      ///                  command line arguments occurs

      Application( const core::string& name,
                   int argc, char *argv[],
                   AutoQuit auto_quit = AutoQuit(true) );

      //------------------------------------------------------------------------
      ///
      /// @param name The name of the application
      /// @param argc The argument count
      /// @param argv An array of arguments
      /// @param additional_options Additional command line options
      /// @param auto_quit Whether the program should be quit if an error in the
      ///                  command line arguments occurs

      Application( const core::string& name,
                   int argc, char *argv[],
                   const options_description& additional_options,
                   AutoQuit auto_quit = AutoQuit(true) );

      //------------------------------------------------------------------------
      ///
      /// @brief Destructor
      ///

      virtual ~Application();

      //------------------------------------------------------------------------
      ///
      /// @brief Getter
      ///

      input::ManagerPtr  getInputManager();
      video::RendererPtr getRenderer();
      scene::ManagerPtr  getSceneManager();

      //------------------------------------------------------------------------
      ///
      /// @brief Process the next frame (input, move, render)
      ///
      /// @return true while the application is running and false if it has
      ///         finished

      bool run();


      double getTimeFromStart();
      double getLastFrameTime();

      void setMaxFPS( unsigned short fps );
      float getFPS();

    protected:

      input::ManagerPtr         _input_manager;
      video::WindowPtr          _window;
      video::RendererPtr        _renderer;
      scene::ManagerPtr         _scene_manager;
      vfs::VirtualFileSystemPtr _vfs;           ///< virtual file system

      utilities::Timer      _timer;
      double                _last_frame_time;
      double                _last_frame_start;

      utilities::DynamicAverage<float,30>  _fps_smooth_calculator;
      utilities::DynamicAverage<double,12> _fps_calculator;
      unsigned short        _max_fps;

      bool                  _run;

      boost::shared_ptr<log::LogFileWriter> _log_file;
      boost::signals::scoped_connection     _log_file_signal;

      //------------------------------------------------------------------------
      // config values

      xml::DocumentPtr  _config_file;
      path              _config_file_path;
      path              _data_path;

    private:

      //------------------------------------------------------------------------
      ///
      /// @brief Parse the command line options
      ///
      /// Parses the command line options given with @p argc and @p argv into
      /// the equivalent static member variables. If the help option is set or
      /// the wrong arguments are given an help message is prompted.
      ///
      /// @warning If an error occurs or the help message is prompted the
      /// program will automatically be exited with a call to std::exit()
      ///
      /// @param argc The argument count
      /// @param argv An array of arguments
      /// @param auto_quit Whether the program should be quit if an error in the
      ///                  command line arguments occurs
      /// @param additional_options Additional command line options

      void
      parseCommandLine( int argc, char *argv[],
                        AutoQuit auto_quit,
                        const options_description *additional_options = NULL );

  };

} // namespace core
} // namespace gewoopi

#endif /* APPLICATION_H_ */
