//------------------------------------------------------------------------------
// file_iterator.hpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 13.10.2009
 */

#ifndef GEWOOPI_FILE_ITERATOR_HPP_
#define GEWOOPI_FILE_ITERATOR_HPP_

#include "ifstream.hpp"

namespace gewoopi
{
namespace core
{
namespace vfs
{

  //----------------------------------------------------------------------------
  ///
  /// @brief file_iterator for a file in the virtual file system.
  ///

  template<class _CharT>
  class file_iterator:
    public boost::iterator_facade
                <
                  file_iterator<_CharT>,
                  _CharT,
                  std::bidirectional_iterator_tag,
                  _CharT,
                  std::ptrdiff_t
                >,
    public utilities::safe_bool<file_iterator<_CharT> >
  {
    public:

      //------------------------------------------------------------------------
      ///
      /// Create an uninitialized iterator
      ///

      file_iterator()
      {}

      //------------------------------------------------------------------------
      ///
      /// Create an file_iterator from the specified file
      ///
      /// @param file_path Path to the file to open

      file_iterator( const path& file_path ):
        _strm ( new ifstream(file_path) ),
        _cur_pos( 0 )
      {}

      //------------------------------------------------------------------------
      ///
      /// Check if the stream this file_iterator points to is ok.
      ///

      bool operator_bool() const
      { return _strm && _strm->operator_bool(); }

      //------------------------------------------------------------------------
      ///
      /// Get the size of the file this file_iterator points to.
      ///
      /// @return Size of the file

      std::streamoff file_size()
      {
        _strm->seekg( 0, std::ios_base::end );
        return _strm->tellg();
      }

      //------------------------------------------------------------------------
      ///
      /// create an file_iterator which points to the end of the stream of this
      /// file_iterator instance
      ///
      /// @return Iterator to the end of this stream.

      file_iterator make_end()
      {
        file_iterator end(*this);
        end._cur_pos = file_size();
        return end;
      }

    private:

      //------------------------------------------------------------------------

      friend class boost::iterator_core_access;

      //------------------------------------------------------------------------
      void increment()
      { ++_cur_pos; }

      void decrement()
      { --_cur_pos; }

      bool equal(file_iterator const& other) const
      { return _strm == other._strm && _cur_pos == other._cur_pos; }

      _CharT dereference() const
      {
        if( _strm->tellg() != _cur_pos )
          _strm->seekg( _cur_pos );

        return _strm->peek();
      }

      //------------------------------------------------------------------------

      boost::shared_ptr<ifstream> _strm;

      std::streamoff    _cur_pos;   ///!< Current position in stream
      bool              _eof;       ///!< End of line reached

  };

} // namespace vfs
} // namespace core
} // namespace gewoopi

#endif /* GEWOOPI_FILE_ITERATOR_HPP_ */
