//------------------------------------------------------------------------------
// virtual_file.hpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 12.10.2009
 */

#ifndef GEWOOPI_VIRTUAL_IFSTREAM_HPP_
#define GEWOOPI_VIRTUAL_IFSTREAM_HPP_

#include <boost/iostreams/stream.hpp>
#include <boost/iostreams/stream_buffer.hpp>
#include <boost/iostreams/categories.hpp>
#include <boost/iostreams/seek.hpp>

#include <iostream>

#include <physfs.h>

#include "../std.hpp"
#include "../../utilities/safe_bool.hpp"

namespace gewoopi
{
namespace core
{
namespace vfs
{
namespace internal
{

  // Tag for #ifstream_device

  struct ifstream_device_tag:
      boost::iostreams::device_tag,
      boost::iostreams::input_seekable,
      boost::iostreams::closable_tag
  {};

  //----------------------------------------------------------------------------
  //
  // This class shouldn't be used directly. See the typedef below
  //

  class ifstream_device
  {
    public:

      //------------------------------------------------------------------------
      // Traits

      typedef char                  char_type;
      typedef ifstream_device_tag   category;

      //------------------------------------------------------------------------
      ///
      /// @brief
      ///

      explicit ifstream_device( const path& file_path );

      //------------------------------------------------------------------------
      ///
      /// Read up to n characters from the underlying data source into the
      /// buffer s
      ///
      /// @return Number of characters read; -1 indicates EOF

      std::streamsize read( char_type* s, std::streamsize n );

      //------------------------------------------------------------------------
      ///
      /// Seek to position off and return the new stream position.
      ///
      /// @param off  Seek offset
      /// @param way  - std::ios_base::beg indicates an offset from the sequence
      ///               beginning
      ///             - std::ios_base::cur indicates an offset from the current
      ///               character position
      ///             - std::ios_base::end indicates an offset from the
      ///               sequence end
      /// @return

      std::streampos seek( std::streamoff off,
                           std::ios_base::seekdir way );

      void close();

    private:

      PHYSFS_File* _file;

  };

} // namespace internal

  //----------------------------------------------------------------------------
  ///
  /// This stream opens a file for read access in the virtual file system
  ///

  class ifstream:
    public boost::iostreams::stream_buffer<internal::ifstream_device>,
    public std::basic_istream<internal::ifstream_device::char_type>,
    public utilities::safe_bool<ifstream>
  {
    public:

      //------------------------------------------------------------------------

      ifstream();

      //------------------------------------------------------------------------
      ///
      /// @brief Open the specified file from the virtual file system
      ///
      /// @param file_path Path to the file to open.

      explicit ifstream( const path& file_path );

      //------------------------------------------------------------------------

      bool operator_bool() const;

  };

} // namespace vfs
} // namespace core
} // namespace gewoopi

#endif /* GEWOOPI_VIRTUAL_IFSTREAM_HPP_ */
