//------------------------------------------------------------------------------
// manager.h - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file: input/manager.h
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 05.07.2009
 */

#ifndef INPUT_MANAGER_H_
#define INPUT_MANAGER_H_

#include <vector>
#include <map>

#include "../core/fwd.h"
#include "fwd.h"
#include "button.h"
#include "device.h"

namespace gewoopi
{
namespace input
{

  class Manager
  {
    public:

      Manager();
      ~Manager();

      //------------------------------------------------------------------------
      ///
      /// @brief Create a new input device
      ///
      /// @param type           The type of the device (Keyboard, Mouse, ...)
      /// @param name           The name of the device
      /// @param num_buttons    The number of buttons this device has
      /// @return Id of the new device

      DeviceId createDevice( Device::Types type,
                             const core::string& name,
                             size_t num_buttons );

      //------------------------------------------------------------------------
      ///
      /// @brief Get the device at position device in the internal device array
      /// @note 0 is always the system keyboard and 1 always the system mouse
      ///
      /// @param device Id of the requested device
      /// @return DevicePtr for the requested device

      DevicePtr getDevice( DeviceId device );

      //------------------------------------------------------------------------
      ///
      /// @brief Get the device which represents the system keyboard
      ///
      /// @return DevicePtr for the system keyboard

      DevicePtr getKeyboard();

      //------------------------------------------------------------------------
      ///
      /// @brief Get the device which represents the system mouse
      ///
      /// @return DevicePtr for the system mouse

      DevicePtr getMouse();

      //------------------------------------------------------------------------
      ///
      /// @brief Used by the input system to update the virtual devices
      ///

      void notifyAbsoluteChange( DeviceId device, size_t button, float value );
      void notifyRelateiveChange( DeviceId device, size_t button, float value );

      //------------------------------------------------------------------------
      ///
      /// @brief Add a button to the internal button map
      ///
      /// @param button The button to add

      void addButton( const Button& button );

      //------------------------------------------------------------------------
      ///
      /// @brief Get the button with the specified name
      ///
      /// If there is no such button specified there is also returned a button,
      /// but it is just a dummy button which always has a value of 0 and equals
      /// to false.
      /// This function should be used to get the virtual buttons specified in
      /// the config file for the button mapping, which is loaded by the
      /// initVirtualButtons(). There is also a possibility to add virtual
      /// buttons with createButton.
      ///
      /// @param name Name of the button
      /// @return The button

      const Button getButton( const core::string& name ) const;

      //------------------------------------------------------------------------
      ///
      /// @brief Initialize the virtual buttons
      ///
      /// Create the mapping between the existing hardware buttons and the
      /// virtual ones.

      void initVirtualButtons();

    private:

      typedef std::vector<DevicePtr> DeviceList;
      typedef std::map<core::string,Button> ButtonMap;

      DeviceList _devices;
      ButtonMap  _virtual_buttons;
  };

} // namespace input
} // namespace gewoopi

#endif /* INPUT_MANAGER_H_ */
