//------------------------------------------------------------------------------
// matrix.h - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#ifndef GEWOOPI_MATRIX_H_
#define GEWOOPI_MATRIX_H_

#include <cmath>
#include <stdexcept>
#include <cstring>
#include <iostream>
#include <iomanip>

#include <gewoopi/utilities/utilities.h>
#include "vector.hpp"
#include "constants.hpp"
#include "angle.hpp"

namespace gewoopi
{
namespace math
{

  //----------------------------------------------------------------------------
  ///
  /// @brief 4x4 Matrix
  ///

  template<typename T> class Matrix4x4
  {
    public:

      //------------------------------------------------------------------------

      union
      {
        T m_[4][4];
        T m_a[16];
        struct
        {
          T m11_, m12_, m13_, m14_,
            m21_, m22_, m23_, m24_,
            m31_, m32_, m33_, m34_,
            m41_, m42_, m43_, m44_;
        };
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Default constructor creates a identity matrix
      ///

      Matrix4x4():
        m11_(1), m12_(0), m13_(0), m14_(0),
        m21_(0), m22_(1), m23_(0), m24_(0),
        m31_(0), m32_(0), m33_(1), m34_(0),
        m41_(0), m42_(0), m43_(0), m44_(1)
      {

      }

      Matrix4x4( T m[4][4] )
      {
        memcpy(m_,m,16*sizeof(T));
      }

      Matrix4x4( T m11, T m12, T m13, T m14,
                 T m21, T m22, T m23, T m24,
                 T m31, T m32, T m33, T m34,
                 T m41, T m42, T m43, T m44 ):
        m11_(m11), m12_(m12), m13_(m13), m14_(m14),
        m21_(m21), m22_(m22), m23_(m23), m24_(m24),
        m31_(m31), m32_(m32), m33_(m33), m34_(m34),
        m41_(m41), m42_(m42), m43_(m43), m44_(m44)
      {

      }

      //------------------------------------------------------------------------
      ///
      /// @brief Get the determinant of this matrix.
      ///
      /// @return Determinant of this matrix.

      T det() const
      {
        // sum of main diagonals minus secondary diagonals
        // (caclulate it just of the 3x3 matrix because we
        // dont need more for our transformations)
        return   m11_ * (m22_ * m33_ - m23_ * m32_)
               - m12_ * (m21_ * m33_ - m23_ * m31_)
               + m13_ * (m21_ * m32_ - m22_ * m31_);
      }

      //------------------------------------------------------------------------
      ///
      /// Get the inverse of this matrix
      ///
      /// Returns the inverse matrix if is regular else the identity matrix.
      ///
      /// @return the resulting 4x4 inverted matrix

      const Matrix4x4<T> inverse() const
      {
        T determinant = det();

        if( determinant == 0 )
          return Identity();

        T inv_det = 1.0 / determinant;

        Matrix4x4<T> ret;

        ret.m11_ =  inv_det * (m22_ * m33_ - m23_ * m32_);
        ret.m12_ = -inv_det * (m12_ * m33_ - m13_ * m32_);
        ret.m13_ =  inv_det * (m12_ * m23_ - m13_ * m22_);
        ret.m14_ =  0;
        ret.m21_ = -inv_det * (m21_ * m33_ - m23_ * m31_);
        ret.m22_ =  inv_det * (m11_ * m33_ - m13_ * m31_);
        ret.m23_ = -inv_det * (m11_ * m23_ - m13_ * m21_);
        ret.m24_ =  0;
        ret.m31_ =  inv_det * (m21_ * m32_ - m22_ * m31_);
        ret.m32_ = -inv_det * (m11_ * m32_ - m12_ * m31_);
        ret.m33_ =  inv_det * (m11_ * m22_ - m12_ * m21_);
        ret.m34_ =  0;
        ret.m41_ = -(m41_ * ret.m11_ + m42_ * ret.m21_ + m43_ * ret.m31_);
        ret.m42_ = -(m41_ * ret.m12_ + m42_ * ret.m22_ + m43_ * ret.m32_);
        ret.m43_ = -(m41_ * ret.m13_ + m42_ * ret.m23_ + m43_ * ret.m33_);
        ret.m44_ = 1;

        return ret;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Access elements with (row,col)
      ///
      /// @param row Row of element
      /// @param col Column of element
      ///
      /// @return Reference to matrix element at (row,col)

      T& operator()( unsigned int row, unsigned int col )
      {
        THROW_CHECK_DEBUG( row > rows || col > cols ||
                           row == 0   || col == 0,
                           std::range_error("Matrix operator()") );
        return m_[row-1][col-1];
      }

      const T& operator()( unsigned int row, unsigned int col ) const
      {
        THROW_CHECK_DEBUG( row > rows || col > cols ||
                           row == 0   || col == 0,
                           std::range_error("Matrix operator()") );
        return m_[row-1][col-1];
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Access elements with [index]
      ///
      /// @param index Zerobased index of the matrix element
      ///
      /// @return Reference to matrix element with specified index

      T& operator[]( unsigned int index )
      {
        THROW_CHECK_DEBUG( index > 15 || index == 0,
                           std::range_error("Matrix operator[]") );
        return m_a[index];
      }

      const T& operator[]( unsigned int index ) const
      {
        THROW_CHECK_DEBUG( index > 15 || index == 0,
                           std::range_error("Matrix operator[]") );
        return m_a[index];
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator+
      ///

      const Matrix4x4 operator+ ( const Matrix4x4<T>& rhs ) const
      {
        return Matrix4x4(
          m11_ + rhs.m11_, m12_ + rhs.m12_, m13_ + rhs.m13_, m14_ + rhs.m14_,
          m21_ + rhs.m21_, m22_ + rhs.m22_, m23_ + rhs.m23_, m24_ + rhs.m24_,
          m31_ + rhs.m31_, m32_ + rhs.m32_, m33_ + rhs.m33_, m34_ + rhs.m34_,
          m41_ + rhs.m41_, m42_ + rhs.m42_, m43_ + rhs.m43_, m44_ + rhs.m44_
                        );
      }

      const Matrix4x4& operator+= ( const Matrix4x4& rhs )
      {
        for( int i = 0; i < 16; ++i )
          m_a[i] += rhs.m_a[i];
        return *this;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator-
      ///

      const Matrix4x4 operator- ( const Matrix4x4<T>& rhs ) const
      {
        return Matrix4x4(
          m11_ - rhs.m11_, m12_ - rhs.m12_, m13_ - rhs.m13_, m14_ - rhs.m14_,
          m21_ - rhs.m21_, m22_ - rhs.m22_, m23_ - rhs.m23_, m24_ - rhs.m24_,
          m31_ - rhs.m31_, m32_ - rhs.m32_, m33_ - rhs.m33_, m34_ - rhs.m34_,
          m41_ - rhs.m41_, m42_ - rhs.m42_, m43_ - rhs.m43_, m44_ - rhs.m44_
                        );
      }

      const Matrix4x4& operator-= ( const Matrix4x4& rhs )
      {
        for( int i = 0; i < 16; ++i )
          m_a[i] -= rhs.m_a[i];
        return *this;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator*
      ///

      const Matrix4x4 operator* ( const Matrix4x4<T>& rhs ) const
      {
        return Matrix4x4( rowMulCol(*this,rhs,1,1),
                          rowMulCol(*this,rhs,1,2),
                          rowMulCol(*this,rhs,1,3),
                          rowMulCol(*this,rhs,1,4),

                          rowMulCol(*this,rhs,2,1),
                          rowMulCol(*this,rhs,2,2),
                          rowMulCol(*this,rhs,2,3),
                          rowMulCol(*this,rhs,2,4),

                          rowMulCol(*this,rhs,3,1),
                          rowMulCol(*this,rhs,3,2),
                          rowMulCol(*this,rhs,3,3),
                          rowMulCol(*this,rhs,3,4),

                          rowMulCol(*this,rhs,4,1),
                          rowMulCol(*this,rhs,4,2),
                          rowMulCol(*this,rhs,4,3),
                          rowMulCol(*this,rhs,4,4) );
      }

      const Matrix4x4<T>& operator*= ( const Matrix4x4<T>& rhs )
      {
        for( int row = 1; row <= 4; ++row )
          for( int col = 1; col <= 4; ++col )
            m_[row-1][col-1] = rowMulCol(*this,rhs,row,col);
        return *this;
      }

      //------------------------------------------------------------------------

      const Matrix4x4<T> operator* ( const T& scalar ) const
      {
        return Matrix4x4<T>(
                 m11_ * scalar, m12_ * scalar, m13_ * scalar, m14_ * scalar,
                 m21_ * scalar, m22_ * scalar, m23_ * scalar, m24_ * scalar,
                 m31_ * scalar, m32_ * scalar, m33_ * scalar, m34_ * scalar,
                 m41_ * scalar, m42_ * scalar, m43_ * scalar, m44_ *scalar );
      }

      const Matrix4x4<T>& operator*= ( const T& scalar )
      {
        for( int i = 0; i < 16; ++i )
          m_[i] *= scalar;
        return *this;
      }

      //------------------------------------------------------------------------

      const Vector3<T> operator* ( const Vector3<T>& rhs ) const
      {
        return Vector3<T>(   m11_ * rhs.value_[0]
                           + m21_ * rhs.value_[1]
                           + m31_ * rhs.value_[2]
                           + m41_,
                             m12_ * rhs.value_[0]
                           + m22_ * rhs.value_[1]
                           + m32_ * rhs.value_[2]
                           + m42_,
                             m13_ * rhs.value_[0]
                           + m23_ * rhs.value_[1]
                           + m33_ * rhs.value_[2]
                           + m43_ );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator==
      ///

      bool operator== ( const Matrix4x4<T>& rhs ) const
      {
        for( int i = 0; i < 16; ++i )
          if( m_a[i] != rhs.m_a[i] ) return false;
        return true;
      }

      //------------------------------------------------------------------------

      operator const T* const() const { return &m11_; };

      //------------------------------------------------------------------------

      //------------------------------------------------------------------------
      ///
      /// @brief Create an identity matrix.
      ///
      /// @return The identity matrix.

      static const Matrix4x4<T> Identity()
      {
        return Matrix4x4<T>( 1,0,0,0,
                             0,1,0,0,
                             0,0,1,0,
                             0,0,0,1 );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Create a left-handed perspective projection matrix
      ///
      /// @return The projection matrix.

      static const Matrix4x4<T> PerspectiveLH( T width,
                                               T height,
                                               T z_near,
                                               T z_far )
      {
        T dist = z_near-z_far;
        if(    dist > -0.00005 || z_near < 0.00005
            || width < 0.00005 || height < 0.00005 )
          return Identity();

        T inv_dist = 1/dist;

        return Matrix4x4<T>(
          2*z_near/width, 0,             0,                    0,
          0,            2*z_near/height, 0,                    0,
          0,            0,             (z_near+z_far)*inv_dist, -1,
          0,            0,             2*z_near*z_far*inv_dist,  0 );
      }

      static const Matrix4x4<T> PerspectiveLH( const Size2D<T>& size,
                                               T z_near,
                                               T z_far )
      {
        return PerspectiveLH( size.width(), size.height(), z_near, z_far );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Create a right-handed perspective projection matrix
      ///
      /// @return The projection matrix.

      static const Matrix4x4<T> PerspectiveRH( T width,
                                               T height,
                                               T z_near,
                                               T z_far )
      {
        T dist = z_near-z_far;
        if(    dist > -0.00005 || z_near < 0.00005
            || width < 0.00005 || height < 0.00005 )
          return Identity();

        T inv_dist = 1/dist;

        return Matrix4x4<T>(
          2*z_near/width, 0,             0,                     0,
          0,            2*z_near/height, 0,                     0,
          0,            0,               z_far*inv_dist,        1,
          0,            0,              -z_near*z_far*inv_dist, 0 );
      }

      static const Matrix4x4<T> OrthoLH( T width, T height, T z_near, T z_far )
      {
        T dist = z_near-z_far;
        if(    dist > -0.00005 || z_near < 0.00005
            || width < 0.00005 || height < 0.00005 )
          return Identity();

        T inv_dist = 1/dist;
        return Matrix4x4<T>( 2/width, 0,        0,                       0,
                             0,       2/height, 0,                       0,
                             0,       0,        2*inv_dist,              0,
                             0,       0,        (z_near+z_far)*inv_dist, 1 );
      }

      static const Matrix4x4<T> OrthoRH( T width, T height, T z_near, T z_far )
      {
        T dist = z_near-z_far;
        if(    dist > -0.00005 || z_near < 0.00005
            || width < 0.00005 || height < 0.00005 )
          return Identity();

        T inv_dist = 1/dist;
        return Matrix4x4<T>( 2/width, 0,        0,               0,
                             0,       2/height, 0,               0,
                             0,       0,        -inv_dist,       0,
                             0,       0,        z_near*inv_dist, 1 );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Create a Translation Matrix
      ///

      static const Matrix4x4<T> Translate( T x, T y, T z )
      {
        return Matrix4x4<T>( 1,0,0,0,
                             0,1,0,0,
                             0,0,1,0,
                             x,y,z,1 );
      }

      static const Matrix4x4<T> Translate( const Vector3<T>& vec )
      {
        return Translate( vec.x(), vec.y(), vec.z() );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Create a Rotation Matrix
      ///

      static const Matrix4x4<T> RotateX( const Radian<T> angle )
      {
        T sin = std::sin(angle);
        T cos = std::cos(angle);

        return Matrix4x4<T>( 1,   0,  0,0,
                             0, cos,sin,0,
                             0,-sin,cos,0,
                             0,   0,  0,1 );
      }

      static const Matrix4x4<T> RotateY( const Radian<T> angle )
      {
        T sin = std::sin(angle);
        T cos = std::cos(angle);

        return Matrix4x4<T>( cos,0,-sin,0,
                             0,  1,   0,0,
                             sin,0, cos,0,
                             0,  0,   0,1 );
      }

      static const Matrix4x4<T> RotateZ( const Radian<T> angle )
      {
        T sin = std::sin(angle);
        T cos = std::cos(angle);

        return Matrix4x4<T>( cos,sin,0,0,
                            -sin,cos,0,0,
                               0,  0,1,0,
                               0,  0,0,1 );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief create a left-handed look-at transformation matrix
      ///
      /// Create a look-at transformation matrix for a left-handed coordinate
      /// system.
      ///
      /// @param pos position where observer looks from
      /// @param at the position the observer should look to
      /// @param up normalized up vector
      ///
      /// @return the resulting 4x4 matrix

      static const Matrix4x4<T> LookAtLH( const Vector3<T>& pos,
                                          const Vector3<T>& at,
                                          const Vector3<T>& up )
      {
        Vector3<T> z_axis( pos - at );
        z_axis.normalize();

        Vector3<T> x_axis( up.cross(z_axis) );
        x_axis.normalize();

        Vector3<T> y_axis( z_axis.cross(x_axis) );

        return Matrix4x4<T>( x_axis.x(),  y_axis.x(),  z_axis.x(), 0,
                             x_axis.y(),  y_axis.y(),  z_axis.y(), 0,
                             x_axis.z(),  y_axis.z(),  z_axis.z(), 0,
                            -x_axis*pos, -y_axis*pos, -z_axis*pos, 1 );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief create a right-handed look-at transformation matrix
      ///
      /// Create a look-at transformation matrix for a right-handed coordinate
      /// system.
      ///
      /// @param pos position where observer looks from
      /// @param at the position the observer should look to
      /// @param up normalized up vector
      ///
      /// @return the resulting 4x4 matrix

      static const Matrix4x4<T> LookAtRH( const Vector3<T>& pos,
                                          const Vector3<T>& at,
                                          const Vector3<T>& up )
      {
        Vector3<T> z_axis( at - pos );
        z_axis.normalize();

        Vector3<T> x_axis( up.cross(z_axis) );
        x_axis.normalize();

        Vector3<T> y_axis( z_axis.cross(x_axis) );

        return Matrix4x4<T>( x_axis.x(),  y_axis.x(),  z_axis.x(), 0,
                             x_axis.y(),  y_axis.y(),  z_axis.y(), 0,
                             x_axis.z(),  y_axis.z(),  z_axis.z(), 0,
                            -x_axis*pos, -y_axis*pos, -z_axis*pos, 1 );
      }

    private:

      T rowMulCol( const Matrix4x4<T>& lhs, const Matrix4x4<T>& rhs,
                   unsigned int row, unsigned int col ) const
      {
        T sum = T();
        for( int i = 0; i < 4; ++i )
          sum += lhs.m_[row-1][i] * rhs.m_[i][col-1];
        return sum;
      }
  };

  template<typename T>
    std::ostream& operator <<( std::ostream& strm, const Matrix4x4<T>& m )
  {
    strm << "\n" << std::fixed << std::setw(9) << std::setprecision(3) << m[0];

    for( unsigned int i = 1; i < 16; ++i )
    {
      strm << " " << std::fixed << std::setw(8) << std::setprecision(3) << m[i];
      if( !((i + 1) % 4) ) strm << std::endl;
    }

    return strm;
  }

  typedef Matrix4x4<float> Matrix4x4f;

} // namespace math
} // namespace gewoopi

#endif /* GEWOOPI_MATRIX_H_ */
