//------------------------------------------------------------------------------
// plane.hpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 16.10.2009
 */

#ifndef GEWOOPI_MATH_PLANE_HPP_
#define GEWOOPI_MATH_PLANE_HPP_

#include "ray.hpp"

namespace gewoopi
{
namespace math
{

  //----------------------------------------------------------------------------
  ///
  /// A plane in 3d space (ax + by + cz = d)
  ///

  class Plane
  {
    public:

      //------------------------------------------------------------------------
      ///
      /// Construct a plane from three points on the plane
      ///

      Plane( const Vector3f& p0, const Vector3f& p1, const Vector3f& p2 );

      //------------------------------------------------------------------------
      ///
      /// Create a plane from a point on the plane and the normal
      ///

      Plane( const Vector3f& point, const Vector3f& normal );

      //------------------------------------------------------------------------
      ///
      /// Create just an empty plane object
      ///

      Plane();

      //------------------------------------------------------------------------
      ///
      /// Get the normal vector of the plane
      ///
      /// @return The normal vector

      const Vector3f& getNormal() const;

      //------------------------------------------------------------------------
      ///
      /// Get the minimal distance of the plane to the origin
      ///
      /// @return The distance to the origin

      real getD() const;

      //------------------------------------------------------------------------
      ///
      /// Get the distance of a point to the plane
      ///
      /// @param vec    A vector
      /// @return The distance of the given vector to the plane


      real distanceTo( const Vector3f& vec ) const;

      //------------------------------------------------------------------------
      ///
      /// Check if a point is on the plane
      ///

      bool contains( const Vector3f& point ) const;

      //------------------------------------------------------------------------
      ///
      /// @brief Check if a point is behind this plane
      ///

      bool isBehind( const Vector3f& point ) const;

      //------------------------------------------------------------------------
      ///
      /// Check for a collision
      ///
      /// @param ray      The ray to check the collision with
      /// @param distance If not zero the distance is stored at it's address
      /// @param normal   If not zero the normal at the hitpoint is stored at
      ///                 it's address

      bool hit( const Ray3f& ray,
                real* distance   = NULL,
                Vector3f* normal = NULL ) const;

    private:

      //ax + by + cz = d

      Vector3f _normal;
      real     _d;

  };

  //----------------------------------------------------------------------------

  typedef Plane Planef; ///< Just make it conformant to the naming scheme

  //----------------------------------------------------------------------------
  std::ostream& operator <<( std::ostream& strm, const Plane& plane );

} // namespace math
} // namespace gewoopi

#endif /* GEWOOPI_MATH_PLANE_HPP_ */
