//------------------------------------------------------------------------------
// geometric_vectors.h - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#ifndef GEWOOPI_MATH_VECTOR_H_
#define GEWOOPI_MATH_VECTOR_H_

#include <cmath>
#include <stdexcept>
#include <iostream>
#include <string>
#include <assert.h>

#include "float_compare.hpp"

namespace gewoopi
{
namespace math
{

  //----------------------------------------------------------------------------
  ///
  /// @brief Base Vector
  ///

  template<typename T, unsigned short dim> class Vector
  {
    public:

      //------------------------------------------------------------------------

      T value_[dim];

      //------------------------------------------------------------------------
      ///
      /// constructor
      ///
      /// @param def_val Default value for all components

      explicit Vector( const T& def_val = T() )
      {
        for( unsigned int i = 0; i < dim; ++i )
          value_[i] = def_val;
      };

      explicit Vector( T val[dim] )
      {
        for( unsigned int i = 0; i < dim; ++i )
          value_[i] = val[i];
      };

      Vector( const Vector<T,dim>& src )
      {
        *this = src;
      }

      const Vector<T,dim>& operator= ( const Vector<T,dim>& src )
      {
        for( unsigned int i = 0; i < dim; ++i )
                  value_[i] = src.value_[i];
        return *this;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Get an the element at position index
      ///
      /// @return A reference to the element

      T& operator[]( unsigned int index )
      {
        if( index >= dim ) throw std::range_error("Vector operator[]");
        return value_[index];
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Get pointer to components
      ///
      /// Returns pointer to the internal array of components
      ///
      /// @return Pointer to the first component

      const T* getPointer() const
      {
        return &value_[0];
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Get the square of the length of the Vector
      ///
      /// This method is meant to be used when you do not need the real length
      /// of the vector but just any value witch is usable to compare with an
      /// other vector. For the sake of speed we can use the square length
      /// in this cases.
      ///
      /// @return Square of vector length

      T lengthSqr() const
      {
        T sum = value_[0] * value_[0];
        for( unsigned int i = 1; i < dim; ++i )
          sum += value_[i] * value_[i];
        return sum;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Get the length of the Vector
      ///
      /// @return Vector length

      T length() const
      {
        return sqrt( lengthSqr() );
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Normalize this vector
      ///
      /// @return The normalized vector

      const Vector<T,dim>& normalize()
      {
        T len = length();
        if( !len ) return *this;

        for( unsigned int i = 0; i < dim; ++i )
          value_[i] /= len;
        return *this;
      };

      const Vector<T,dim> normalize() const
      {
        T len = length();
        if( !len ) return *this;
        return *this*(1./len);
      };

      //------------------------------------------------------------------------
      ///
      /// @brief
      ///
      /// @return description of return value

      T distance( const Vector<T,dim>& second ) const
      {
        T sqr_sum( (T) square(second.value_[0] - value_[0]) );
        for( unsigned int i = 1; i < dim; ++i )
          sqr_sum += square( second.value_[i] - value_[i] );
        return sqrt( sqr_sum );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Operator+=
      ///
      /// @return description of return value

      Vector<T,dim>& operator+= ( const Vector<T,dim>& rhs )
      {
        for( unsigned int i = 0; i < dim; ++i )
          value_[i] += rhs.value_[i];
        return *this;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Operator+
      ///
      /// @return description of return value

      inline Vector<T,dim> operator+ ( const Vector<T,dim>& rhs ) const
      {
        Vector<T,dim> ret( *this );

        return ret += rhs;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Operator-=
      ///
      /// @return description of return value

      const Vector<T,dim>& operator-= ( const Vector<T,dim>& rhs )
      {
        for( unsigned int i = 0; i < dim; ++i )
          value_[i] -= rhs.value_[i];
        return *this;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Operator-
      ///
      /// @return description of return value

      const Vector<T,dim> operator- ( const Vector<T,dim>& rhs ) const
      {
        Vector<T,dim> ret( *this );

        return ret -= rhs;
      };

      const Vector<T,dim> operator- () const
      {
        Vector<T,dim> ret( *this );

        for( unsigned int i = 0; i < dim; ++i )
          ret.value_[i] = -value_[i];

        return ret;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Operator*=
      ///
      /// @return description of return value

      const Vector<T,dim>& operator*= ( const T scal )
      {
        for( unsigned int i = 0; i < dim; ++i )
          value_[i] *= scal;

        return *this;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator*
      ///
      /// @return description of return value

      const Vector<T,dim> operator * ( const T scal ) const
      {
        return Vector<T,dim>(*this) *= scal;
      }

      const T operator * ( const Vector<T,dim>& rhs ) const
      {
        T ret( value_[0] * rhs.value_[0] );
        for( unsigned int i = 1; i < dim; ++i )
          ret += value_[i] * rhs.value_[i];
        return ret;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief operator ==
      ///
      /// Compare two Vectors

      template<typename T_other>
        bool operator ==( const Vector<T_other,dim>& rhs ) const
      {
        for( unsigned short i = 0; i < dim; ++i )
        {
          if( !isEqual(value_[i], rhs.value_[i]) )
            return false;
        }
        return true;
      }

      bool operator ==( const Vector<T,dim>& rhs ) const
      {
        return operator == <T>(rhs);
      }

      //------------------------------------------------------------------------
      ///

      operator const T* const () const
      {
        return &value_[0];
      }

  };

  //----------------------------------------------------------------------------
  ///
  /// @brief Protect against use of Vector without components
  ///

  template<typename T> class Vector<T,0>
  {
    Vector(); // private, because a vector
              // without components is not allowed
  };

  //----------------------------------------------------------------------------
  ///
  /// @brief 2D Vector
  ///

  template<typename T> class Vector2: public Vector<T,2>
  {
    public:

      explicit Vector2( const T& def_val = T() ): Vector<T,2>::Vector(def_val)
      {};

      Vector2( T val[2] ): Vector<T,2>::Vector(val)
      {};

      Vector2( const T x, const T y )
      {
        Vector<T,2>::value_[0] = x;
        Vector<T,2>::value_[1] = y;
      };

      /*const Vector2<T> operator =( const Vector<T,2>& src )
      {
        return *static_cast<Vector<T,2>*>(this) = src;
      }*/

      //------------------------------------------------------------------------
      ///
      /// @brief Set components
      ///

      void x( T val ) { Vector<T,2>::value_[0] = val; };
      void y( T val ) { Vector<T,2>::value_[1] = val; };

      void u( T val ) { Vector<T,2>::value_[0] = val; };
      void v( T val ) { Vector<T,2>::value_[1] = val; };

      //------------------------------------------------------------------------
      ///
      /// @brief Get components
      ///

      T x() const { return Vector<T,2>::value_[0]; };
      T y() const { return Vector<T,2>::value_[1]; };

      T u() const { return Vector<T,2>::value_[0]; };
      T v() const { return Vector<T,2>::value_[1]; };
  };

  //----------------------------------------------------------------------------
  ///
  /// @brief 3D Vector
  ///

  template<typename T> class Vector3: public Vector<T,3>
  {
    public:

      explicit Vector3( const T& def_val = T() ): Vector<T,3>::Vector(def_val)
      {};

      explicit Vector3( T val[3] ): Vector<T,3>::Vector(val)
      {};

      //Vector3(){};

      Vector3( const T x, const T y, const T z )
      {
        Vector<T,3>::value_[0] = x;
        Vector<T,3>::value_[1] = y;
        Vector<T,3>::value_[2] = z;
      };

      Vector3( const Vector<T,3>& src )
      {
        for( unsigned int i = 0; i < 3; ++i )
          Vector<T,3>::value_[i] = src.value_[i];
      }

      /*const Vector3<T> operator =( const Vector<T,3>& src )
      {
        return *static_cast<Vector<T,3>*>(this) = src;
      }*/

      //------------------------------------------------------------------------
      ///
      /// @brief cross product
      ///

      const Vector3<T> cross( const Vector3<T>& rhs ) const
      {
        return Vector3<T>(
          (*this)[1] * rhs[2] - (*this)[2] * rhs[1],
          (*this)[2] * rhs[0] - (*this)[0] * rhs[2],
          (*this)[0] * rhs[1] - (*this)[1] * rhs[0] );
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Set components
      ///

      void x( T val ) { Vector<T,3>::value_[0] = val; };
      void y( T val ) { Vector<T,3>::value_[1] = val; };
      void z( T val ) { Vector<T,3>::value_[2] = val; };

      //------------------------------------------------------------------------
      ///
      /// @brief Get components
      ///

      T x() const { return Vector<T,3>::value_[0]; };
      T y() const { return Vector<T,3>::value_[1]; };
      T z() const { return Vector<T,3>::value_[2]; };

      T& x() { return Vector<T,3>::value_[0]; };
      T& y() { return Vector<T,3>::value_[1]; };
      T& z() { return Vector<T,3>::value_[2]; };
  };

  //----------------------------------------------------------------------------
  ///
  /// @brief RGB-Color
  ///

  template<typename T> class Color3: public Vector<T,3>
  {
    public:

      Color3( const T r, const T g, const T b )
      {
        Vector<T,3>::value_[0] = r;
        Vector<T,3>::value_[1] = g;
        Vector<T,3>::value_[2] = b;
      };

      Color3(){};

      Color3( const Vector<T,3>& src )
      {
        *static_cast<Vector<T,3>*>(this) = src;
      }

      //------------------------------------------------------------------------
      ///
      /// @brief Set components
      ///

      void r( T val ) { Vector<T,3>::value_[0] = val; };
      void g( T val ) { Vector<T,3>::value_[1] = val; };
      void b( T val ) { Vector<T,3>::value_[2] = val; };

      //------------------------------------------------------------------------
      ///
      /// @brief Get components
      ///

      T r() const { return Vector<T,3>::value_[0]; };
      T g() const { return Vector<T,3>::value_[1]; };
      T b() const { return Vector<T,3>::value_[2]; };

      //------------------------------------------------------------------------

      const Color3<T> operator * ( const Color3<T>& rhs ) const
      {
        return Color3<T>( rhs.value_[0] * Vector<T,3>::value_[0],
                          rhs.value_[1] * Vector<T,3>::value_[1],
                          rhs.value_[2] * Vector<T,3>::value_[2] );
      }

      //------------------------------------------------------------------------

      static Color3<T> const BLACK;
      static Color3<T> const RED;
      static Color3<T> const GREEN;
      static Color3<T> const BLUE;
      static Color3<T> const YELLOW;
      static Color3<T> const ORANGE;
      static Color3<T> const WHITE;

  };

  template<typename T> Color3<T> const Color3<T>::BLACK (0,  0,0);
  template<typename T> Color3<T> const Color3<T>::RED   (1,  0,0);
  template<typename T> Color3<T> const Color3<T>::GREEN (0,  1,0);
  template<typename T> Color3<T> const Color3<T>::BLUE  (0,  0,1);
  template<typename T> Color3<T> const Color3<T>::YELLOW(1,  1,0);
  template<typename T> Color3<T> const Color3<T>::ORANGE(1,0.5,0);
  template<typename T> Color3<T> const Color3<T>::WHITE (1,  1,1);

  //----------------------------------------------------------------------------
  ///
  /// @brief RGBA-Color
  ///

  template<typename T> class Color4: public Vector<T,4>
  {
    public:

      Color4(){};

      Color4( const T r, const T g, const T b, const T a )
      {
        Vector<T,4>::value_[0] = r;
        Vector<T,4>::value_[1] = g;
        Vector<T,4>::value_[2] = b;
        Vector<T,4>::value_[3] = a;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Set components
      ///

      void r( T val ) { Vector<T,4>::value_[0] = val; };
      void g( T val ) { Vector<T,4>::value_[1] = val; };
      void b( T val ) { Vector<T,4>::value_[2] = val; };
      void a( T val ) { Vector<T,4>::value_[3] = val; };

      //------------------------------------------------------------------------
      ///
      /// @brief Get components
      ///

      T r() const { return Vector<T,4>::value_[0]; };
      T g() const { return Vector<T,4>::value_[1]; };
      T b() const { return Vector<T,4>::value_[2]; };
      T a() const { return Vector<T,4>::value_[3]; };
  };

  //----------------------------------------------------------------------------
  ///
  /// @brief 2D-Size
  ///

  template<typename T> class Size2D: public Vector<T,2>
  {
    public:

      Size2D(){};

      Size2D( const T width, const T height )
      {
        Vector<T,2>::value_[0] = width;
        Vector<T,2>::value_[1] = height;
      };

      //------------------------------------------------------------------------
      ///
      /// @brief Set components
      ///

      void width(  T val ) { Vector<T,2>::value_[0] = val; };
      void height( T val ) { Vector<T,2>::value_[1] = val; };

      //------------------------------------------------------------------------
      ///
      /// @brief Get components
      ///

      T width()  const { return Vector<T,2>::value_[0]; };
      T height() const { return Vector<T,2>::value_[1]; };
  };

  template<typename T, unsigned short dim>
    inline const Vector<T,dim> operator *( const T              scal,
                                           const Vector<T,dim>& vec )
  {
    return vec * scal;
  }

  typedef Vector2<real> Vector2f;
  typedef Vector3<real> Vector3f;

  typedef Color3<real> Color3f;
  typedef Color4<real> Color4f;

  template<typename T, unsigned short dim>
    std::ostream& operator <<( std::ostream& strm, const Vector<T,dim>& vec )
  {
    strm << "(" << vec.value_[0];

    for( unsigned int i = 1; i < dim; ++i )
      strm << "|" << vec.value_[i];

    strm << ")";
    return strm;
  }

} // namespace math
} // namespace gewoopi

#endif /* GEWOOPI_MATH_VECTOR */
