//------------------------------------------------------------------------------
// camera.h - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#ifndef CAMERA_H_
#define CAMERA_H_

#include "../math/matrix.hpp"
#include "fwd.h"

namespace gewoopi
{
namespace scene
{

  class Camera
  {
    public:

      //------------------------------------------------------------------------
      ///
      /// @brief Create a new camera from an projection matrix
      ///
      /// @param projection The projection matrix
      /// @param transform The transform matrix

      Camera( const math::Matrix4x4f& projection =
                math::Matrix4x4f::Identity(),
              const math::Matrix4x4f& transform =
                math::Matrix4x4f::Identity() );

      //------------------------------------------------------------------------
      ///
      /// @brief Destructor
      ///

      virtual ~Camera();

      //------------------------------------------------------------------------
      ///
      /// @brief Get the projection matrix of this camera
      ///
      /// @return The current projection matrix

      const math::Matrix4x4f& getProjection() const;

      //------------------------------------------------------------------------
      ///
      /// @brief Set the transformation matrix of this camera
      ///
      /// @param transform The new transformation matrix

      void setTransform( const math::Matrix4x4f& transform );

      //------------------------------------------------------------------------
      ///
      /// @brief Get the transformation matrix of this camera
      ///
      /// @return The current transformation matrix

      const math::Matrix4x4f& getTransform() const;

    protected:

      mutable math::Matrix4x4f _projection;
      mutable bool _changed; ///< Marks that something hast changed so that the
                             ///  matrix Camera::_projection should be updated
                             ///  before it is read.

      math::Matrix4x4f _transform;

      //------------------------------------------------------------------------
      ///
      /// @brief Update the projection matrix
      ///
      /// @note This method should be implemented by all derived classes so that
      ///       it just calculates the new projection matrix if this method is
      ///       called.

      virtual void update() const;

  };

} // namespace scene
} // namespace gewoopi

#endif /* CAMERA_H_ */
