//------------------------------------------------------------------------------
// camera_perspective.h - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file: camera_perspective.h
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 18.07.2009
 */

#ifndef CAMERA_PERSPECTIVE_H_
#define CAMERA_PERSPECTIVE_H_

#include "camera.h"

namespace gewoopi
{
namespace scene
{

  class CameraPerspective: public Camera
  {
    public:

      //------------------------------------------------------------------------
      ///
      /// @brief Create a new perspective camera
      ///
      /// Construct the projection from the size of the near clipping plane and
      /// the distances of the clipping planes.
      ///
      /// @param size Size of the near clipping plane
      /// @param near Distance of near clipping plane
      /// @param far Distance of far clipping plane

      CameraPerspective( const math::Size2D<math::real>& size,
                         const math::real z_near,
                         const math::real z_far );

      //------------------------------------------------------------------------
      ///
      /// @brief Create a new perspective camera
      ///
      /// Construct the projection from the field of view and the aspect ratio
      /// and the distances of the clipping planes.
      ///
      /// @param fovy Field of view in direction of the x-axis (in degree)
      /// @param aspect Aspect ratio of the clipping planes
      /// @param near Distance of near clipping plane
      /// @param far Distance of far clipping plane

      CameraPerspective( const math::Radian<math::real> fovy,
                         const math::real aspect,
                         const math::real z_near,
                         const math::real z_far );

      //------------------------------------------------------------------------
      ///
      /// @brief Destructor
      ///

      virtual ~CameraPerspective();

      //------------------------------------------------------------------------
      ///
      /// @brief Set the field of view.
      ///
      /// @param fovy The field new field of view (in y-direction)

      void setFOV( float fovy );

      //------------------------------------------------------------------------
      ///
      /// @brief Get the current field of view.
      ///
      /// @return The current field of view.

      const float getFOV() const;

    protected:

      math::Size2D<float>   _size; ///< Size of near clipping plane.
      float                 _near; ///< Distance of near clipping plane.
      float                 _far;  ///< Distance of far clipping plane.

      float _aspect; ///< The aspect ratio of the clipping planes (Is intended
                     ///  for internal use only).

      //------------------------------------------------------------------------
      ///
      /// @brief Update projection matrix
      ///

      void update() const;

  };

} // namespace scene
} // namespace gewoopi

#endif /* CAMERA_PERSPECTIVE_H_ */
