//------------------------------------------------------------------------------
// safe_bool.hpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 12.10.2009
 */

#ifndef SAFE_BOOL_HPP_
#define SAFE_BOOL_HPP_

namespace gewoopi
{
namespace utilities
{

  //----------------------------------------------------------------------------
  ///
  /// A base class to enable a bool comparison according to the safe bool idiom.
  /// To use it you have to derive from this class by the CRTP pattern as
  /// following:
  /// @code
  /// class MyClass: public safe_bool<MyClass>
  /// {
  ///   public:
  ///     bool operator_bool() const
  ///     { return true; /* Implement your one routine */ }
  /// }
  /// @endcode

  template<typename Derived>
  class safe_bool
  {
    public:

      typedef void (safe_bool<Derived>::*bool_type)() const;

      //------------------------------------------------------------------------
      operator bool_type() const
      {
        return   static_cast<const Derived*>(this)->operator_bool()
               ? &safe_bool<Derived>::this_type_does_not_support_comparisons
               : 0;
      }

    protected:

      void this_type_does_not_support_comparisons() const
      {}

      //------------------------------------------------------------------------
      ///
      /// Don't allow direct use. You have do derive from safe_bool.
      ///

      safe_bool()
      {}
  };

  // TODO find a better solution for disabling the comparison operators only if
  // not needed

  //----------------------------------------------------------------------------
  ///
  /// Don't allow operator==
  ///

  /*template<typename T1, typename T2>
  bool operator==( const safe_bool<T1>& lhs, const safe_bool<T2>& rhs )
  {
    lhs.this_type_does_not_support_comparisons();
    return false;
  }

  //----------------------------------------------------------------------------
  ///
  /// Don't allow operator!=
  ///

  template<typename T1, typename T2>
  bool operator!=( const safe_bool<T1>& lhs, const safe_bool<T2>& rhs )
  {
    lhs.this_type_does_not_support_comparisons();
    return false;
  }*/

} // namespace utilities
} // namespace gewoopi

#endif /* SAFE_BOOL_HPP_ */
