//------------------------------------------------------------------------------
// renderer.h - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#ifndef RENDERER_H_
#define RENDERER_H_

#include "../math/math.h"
#include "../scene/fwd.h"
#include "../scene/mesh.hpp"
#include "vertex.h"
#include "material.hpp"
#include "render_operation.h"

namespace gewoopi
{
namespace video
{

  class Renderer
  {
    public:

      //------------------------------------------------------------------------
      ///
      /// Constructor
      ///

      Renderer();
      virtual ~Renderer();

      //------------------------------------------------------------------------
      ///
      /// @brief set new render target
      ///
      /// @return previous render target

      virtual RenderTarget* setRenderTarget( RenderTarget *target );

      //------------------------------------------------------------------------
      ///
      /// @brief set new clear color
      ///
      /// @return previous clear color

      virtual math::Color3f setClearColor( const math::Color3f& color );

      //------------------------------------------------------------------------
      ///
      /// @brief Set the camera to be used for the next rendering oparations
      ///
      /// @param camera The camera.

      void setCamera( scene::CameraPtr camera );

      //------------------------------------------------------------------------
      ///
      /// @brief Load a matrix onto the stack
      ///
      /// By loading a matrix on a stack the topmost matrix gets replaced by the
      /// loaded matrix.
      ///
      /// @param mat The new matrix

      virtual void loadMatrix( const math::Matrix4x4f& mat ) = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief Multiply a matrix onto the stack
      ///
      /// Multiply the topmost matrix on the stack with the given matrix and
      /// replace the old matrix with the result.
      ///
      /// @param mat The new matrix

      virtual void multMatrix( const math::Matrix4x4f& mat ) = 0;

      virtual void pushMatrix() = 0;
      virtual void popMatrix() = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief Must be called once before rendering
      ///

      virtual void beginScene( bool clear_back_buffer = true,
                               bool clear_z_buffer    = true ) = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief Must be called once after finished rendering
      ///

      virtual void endScene() = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief Performs a render operation
      ///
      /// Performs the operation stored in the operation object.
      ///
      /// @param operation Reference to the operation object

      virtual void performOperation( RenderOperation& operation ) = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief
      ///

      virtual void drawPrimitives(        const PrimitiveType  type,
                                          scene::VertexArray vertices,
                                          scene::Count count ) = 0;

      virtual void drawIndexedPrimitives( const PrimitiveType  type,
                                          scene::VertexArray vertices,
                                          scene::Count count,
                                          const math::Color3f *color = 0 ) = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief
      ///

      virtual void setCullMode( const CullMode mode ) = 0;

      //------------------------------------------------------------------------
      ///
      /// @brief
      ///

      virtual void setMaterial( const Material& mat ) = 0;

    protected:

      RenderTarget  *_render_target;
      math::Color3f  _clear_color;

      scene::CameraPtr _camera;

      //------------------------------------------------------------------------

  };

} // namespace gewoopi
} // namespace video

#endif /* RENDERER_H_ */
