//------------------------------------------------------------------------------
// application.cpp - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#include <xml-sipa/element.h>
#include <xml-sipa/iterator.h>

#include <iostream>
#include <iterator>
#include <cmath>

#include "core/application.h"
#include "core/vfs/ifstream.hpp"
#include "input/manager.h"
#include "video/OpenGL/window.h"
#include "video/OpenGL/renderer.h"
#include "scene/manager.h"

namespace gewoopi
{
namespace core
{

  //----------------------------------------------------------------------------
  Application::Application( const core::string& name,
                            int argc, char *argv[],
                            AutoQuit auto_quit ):
    _timer( "CoreTimer" ),
    _last_frame_time( 1/80.f ),
    _last_frame_start( _timer.getElapsedTime() - _last_frame_time ),
    _fps_smooth_calculator( _last_frame_time ),
    _fps_calculator( _last_frame_time ),
    _max_fps( 80 ),
    _run( true ),
    _config_file_path("gewoopi.xml")
  {
    //--------------------------------------------------------------------------
    // parse commandline

    parseCommandLine(argc, argv, auto_quit);
    vfs::VirtualFileSystem::init( argv[0] );

    //--------------------------------------------------------------------------
    // initialize log engine and lets redirect the output to a file

    LOG_INIT();
    _log_file.reset( new log::LogFileWriter(name) );
    _log_file_signal = LOG_CONNECT_WRITER(_log_file);

    // strip path from log entries (eg. needed if compiling with cmake)
#ifdef __LOG_STRIP_PATH__
    LOG_STRIP_PATH(__LOG_STRIP_PATH__);
#endif

    //--------------------------------------------------------------------------
    // read the configuration

    _config_file_path =
      vfs::VirtualFileSystem::instance()->getBaseDir() / _config_file_path;

    try
    {
      _config_file.reset( new xml::Document(_config_file_path.file_string()) );
    }
    catch( const std::exception& ex )
    {
      std::cerr << ex.what() << std::endl;

      // create a new config file because none was found or it was invalid
      _config_file.reset( new xml::Document );
      xml::ElementIterator root( _config_file->root() );
      root->setName("gewoopi-config");

      // set default config values
      root->addAttribute( "res-x",      1024    );
      root->addAttribute( "res-y",      768     );
      root->addAttribute( "fullscreen", 0       );
      root.append_element("vfs-settings")
          .append_element("path")->setText( _data_path.directory_string() );

      _config_file->write( _config_file_path.file_string() );
    }

    xml::ElementIterator root = _config_file->root();

    math::Size2D<int> window_size( root->getAttribute<int>("res-x"),
                                   root->getAttribute<int>("res-y") );

    LOG_FMT_INFO( "Resolution: %ix%i",
                  window_size.width() % window_size.height() );

    //--------------------------------------------------------------------------
    // setup virtual filesystem

    _vfs = vfs::VirtualFileSystem::instance();
    _vfs->mount(_data_path, "/data");

    // check for the write directory in the users home directory
    path write_dir_base = _vfs->getUserDir() / ".gewoopi";
    path write_dir      = write_dir_base / name;

    if( boost::filesystem::exists( write_dir ) )
    {
      if( !boost::filesystem::is_directory( write_dir ) )
        throw std::runtime_error( "Write dir not writable." );
    }
    else
    {
      boost::filesystem::create_directory( write_dir_base );
      if( !boost::filesystem::create_directory( write_dir ) )
        throw std::runtime_error( "Failed to create write directory" );
    }

    _vfs->setWriteDir( write_dir );

    //--------------------------------------------------------------------------
    // prepare the subsystems

    _input_manager.reset( new input::Manager() );
    _renderer.reset( new video::OGLRenderer() );
    _scene_manager.reset( new scene::Manager() );

    // create the window
    _window.reset( new video::OGLWindow( window_size,
                                         video::R8G8B8,
                                         _input_manager ) );

    _window->setCaption("TomProgs - Gewoopi - Demo");
    _window->load();

    _input_manager->initVirtualButtons();

    _scene_manager->setRenderer( _renderer );
  }

  //----------------------------------------------------------------------------
  Application::~Application()
  {
    // save the config file...
    _config_file->write( _config_file_path.file_string() );
    LOG_EXIT();
  }

  //----------------------------------------------------------------------------
  input::ManagerPtr Application::getInputManager()
  {
    return _input_manager;
  }

  //----------------------------------------------------------------------------
  video::RendererPtr Application::getRenderer()
  {
    return _renderer;
  }

  //----------------------------------------------------------------------------
  scene::ManagerPtr Application::getSceneManager()
  {
    return _scene_manager;
  }

  //----------------------------------------------------------------------------
  bool Application::run()
  {
    _renderer->endScene();
    _window->processMessages();

    if( !_window->isAlive() )
      _run = false;

    // calculate duration of last frame
    _last_frame_time  = _timer.getElapsedTime() - _last_frame_start;
    _last_frame_start = _timer.getElapsedTime();

    _fps_smooth_calculator.addValue( _last_frame_time );
    _fps_calculator       .addValue( _last_frame_time );

    if( _max_fps && getLastFrameTime() < 1./_max_fps )
      // the factor was just got experimental
      utilities::sleep( (1./_max_fps - getLastFrameTime()) * 1.96 );

    _renderer->beginScene();
    _scene_manager->render( _timer.getElapsedTime(), getLastFrameTime() );
    _scene_manager->move(   _timer.getElapsedTime(), getLastFrameTime() );

    return _run;
  }

  //----------------------------------------------------------------------------
  double Application::getTimeFromStart()
  {
    return _timer.getElapsedTime();
  }

  //----------------------------------------------------------------------------
  double Application::getLastFrameTime()
  {
    return _fps_calculator.getAverage();
  }

  //----------------------------------------------------------------------------
  void Application::setMaxFPS( unsigned short fps )
  {
    _max_fps = fps;
  }

  //----------------------------------------------------------------------------
  float Application::getFPS()
  {
    if( _fps_smooth_calculator.getAverage() > 0.0001f )
      return 1.f/_fps_smooth_calculator.getAverage();
    else return 0;
  }

  //----------------------------------------------------------------------------
  void
  Application::parseCommandLine( int argc, char *argv[],
                                 AutoQuit auto_quit,
                                 const options_description *additional_options )
  {
    namespace po = boost::program_options;

    // set data path default value to directory of executable
    _data_path = path( argv[0] ).parent_path();

    // declare all allowed options
    po::options_description option_desc("Allowed options");

    option_desc.add_options()
      ( "help", "produces this help message" )
      ( "data-dir,d",
        po::value<path>(&_data_path),
        "path to the data directory" );

    // check for additional options
    if( additional_options )
    {
      option_desc.add( *additional_options );
    }

    po::variables_map variables;
    bool parse_error = false;

    // parse the parameters
    try
    {
      po::store( po::parse_command_line(argc,argv,option_desc), variables);
      po::notify( variables );
    }
    catch( po::error& err )
    {
      // if we shouldn't quit automatically just throw on the exception
      if( !auto_quit ) throw err;

      parse_error = true;
      std::cout << err.what() << std::endl;
    }

    // print help message and exit on error or help option
    if( parse_error || variables.count("help") )
    {
      std::cout << option_desc << std::endl;

      if( auto_quit )
      {
        // just return an error if a parsing error occured
        std::exit( parse_error ? EXIT_FAILURE : EXIT_SUCCESS );
      }
    }
  }

} // namespace core
} // namespace gewoopi
