//------------------------------------------------------------------------------
// ifstream.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 12.10.2009
 */

#include "core/vfs/ifstream.hpp"

namespace gewoopi
{
namespace core
{
namespace vfs
{
namespace internal
{

  //----------------------------------------------------------------------------
  ifstream_device::ifstream_device( const path& file_path )
  {
    _file = PHYSFS_openRead( file_path.string().c_str() );

    if( !_file )
      throw std::runtime_error( "Failed to open " + file_path.string() );
  }

  //----------------------------------------------------------------------------
  std::streamsize ifstream_device::read( char_type* s, std::streamsize n )
  {
    PHYSFS_sint64 bytes_read = PHYSFS_read( _file, s, sizeof(char_type),
                                            static_cast<PHYSFS_uint32>(n) );

    if( bytes_read <= 0 )
    {
      if( PHYSFS_eof(_file) )
        return -1;
      else
        throw std::runtime_error(   std::string("[PhysFS] ")
                                  + PHYSFS_getLastError() );
    }

    return static_cast<std::streamsize>(bytes_read);
  }

  //----------------------------------------------------------------------------
  std::streampos ifstream_device::seek( std::streamoff off,
                                        std::ios_base::seekdir way )
  {
    PHYSFS_sint64 new_pos( off );

    switch( way )
    {
      case std::ios_base::beg:
      {
        // no need to change new_pos
        break;
      }
      case std::ios_base::cur:
      {
        PHYSFS_sint64 cur_pos = PHYSFS_tell(_file);
        if( cur_pos < 0 )
          throw std::runtime_error(   std::string("[PhysFS] ")
                                    + PHYSFS_getLastError() );
        new_pos = cur_pos + static_cast<PHYSFS_sint64>(off);
        break;
      }
      case std::ios_base::end:
      {
        PHYSFS_sint64 end = PHYSFS_fileLength(_file);
        if( end < 0 )
          throw std::runtime_error(
            "Could not determine file size, possibly streamed.");
        new_pos = end - static_cast<PHYSFS_sint64>(off);
        break;
      }
      default:
      {
        throw std::invalid_argument("ifstream_device::seek -> way");
      }
    }

    // check if we should correct
    if( new_pos < 0 )
      new_pos = 0;

    if( !PHYSFS_seek( _file, static_cast<PHYSFS_uint64>(new_pos) ) )
      throw std::runtime_error(   std::string("[PhysFS] ")
                                + PHYSFS_getLastError() );

    return std::streampos( static_cast<PHYSFS_uint32>(new_pos) );
  }

  //----------------------------------------------------------------------------
  void ifstream_device::close()
  {
    if( !PHYSFS_close( _file ) )
      throw std::runtime_error(   std::string("[PhysFS] ")
                                + PHYSFS_getLastError() );
  }

} // namespace internal

  //----------------------------------------------------------------------------
  ifstream::ifstream()
  {

  }

  //----------------------------------------------------------------------------
  ifstream::ifstream( const path& file_path ):
    boost::iostreams::stream_buffer<internal::ifstream_device>( file_path ),
    std::basic_istream<internal::ifstream_device::char_type>  ( this )
  {

  }

  //----------------------------------------------------------------------------
  bool ifstream::operator_bool() const
  {
    return this->good();
  }

} // namespace vfs
} // namespace core
} // namespace gewoopi
