//------------------------------------------------------------------------------
// virtual_file_system.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 11.10.2009
 */

#include <iostream>
#include <physfs.h>
#include "core/vfs/virtual_file_system.hpp"

// Check for the correct PhysFS version because the current FindPhysFS of
// CMake is not able to do this. TODO maybe write a own FindPhyFS module...
#if PHYSFS_VER_MAJOR < 2
#  error "Minimal required Version of PhyFS is 2.0.0"
#  if PHYSFS_VER_MAJOR == 0
#    error "You have a PhysFS installation from the 0.x.x series."
#  else
#    error "You have a PhysFS installation from the 1.x.x series."
#  endif
#endif

namespace gewoopi
{
namespace core
{
namespace vfs
{

  //----------------------------------------------------------------------------
  // Use to call PhysFS functions. It checks the return value and throws an
  // appropriate exception on errors.

  void physfsCall( int ret_val )
  {
    if( ret_val == 0 )
      throw std::runtime_error(   std::string("[PhysFS] ")
                                + PHYSFS_getLastError() );
  }

  //----------------------------------------------------------------------------
  // VirtualFileSystem

  // static singleton ptr
  VirtualFileSystemPtr VirtualFileSystem::_singleton;

  //----------------------------------------------------------------------------
  VirtualFileSystemPtr VirtualFileSystem::instance()
  {
    if( !_singleton )
      throw std::runtime_error( "VirtualFileSystem wasn't initialized yet." );
    return _singleton;
  }

  //----------------------------------------------------------------------------
  void VirtualFileSystem::init( const char* argv0 )
  {
    _singleton.reset( new VirtualFileSystem(argv0) );
  }

  //----------------------------------------------------------------------------
  VirtualFileSystem::~VirtualFileSystem()
  {
    physfsCall(PHYSFS_deinit());
  }

  //----------------------------------------------------------------------------
  void VirtualFileSystem::mount( const path& real_path,
                                 const core::string& mount_point )
  {
    physfsCall(PHYSFS_mount( real_path.file_string().c_str(),
                             mount_point.c_str(),
                             1 ));
  }

  //----------------------------------------------------------------------------
  void VirtualFileSystem::setWriteDir( const path& write_path )
  {
    physfsCall(PHYSFS_setWriteDir( write_path.directory_string().c_str() ));
  }

  //----------------------------------------------------------------------------
  bool VirtualFileSystem::exists( const core::string& check_path ) const
  {
    return PHYSFS_exists( check_path.c_str() );
  }

  //----------------------------------------------------------------------------
  const path VirtualFileSystem::getUserDir() const
  {
    return path(PHYSFS_getUserDir());
  }

  //----------------------------------------------------------------------------
  const path VirtualFileSystem::getBaseDir() const
  {
    return path(PHYSFS_getBaseDir());
  }

  //----------------------------------------------------------------------------
  VirtualFileSystem::VirtualFileSystem( const char* argv0 )
  {
    physfsCall(PHYSFS_init(argv0));
  }

} // namespace vfs
} // namespace core
} // namespace gewoopi
