//------------------------------------------------------------------------------
// device.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file: device.cpp
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 06.07.2009
 */

#include "input/device.h"
#include "input/physical_button.hpp"
#include "utilities/utilities.h"

#include <boost/foreach.hpp>

namespace gewoopi
{
namespace input
{

  //----------------------------------------------------------------------------
  // Button name arrays
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  ///
  /// Null-terminated array of keyboard button names
  ///

  const char* button_names_keyboard[] =
  {
    "UNKNOWN",
    "BACKSPACE",
    "TAB",
    "CLEAR",
    "RETURN",
    "PAUSE",
    "ESCAPE",
    "SPACE",
    "EXCLAIM",
    "QUOTEDBL",
    "HASH",
    "DOLLAR",
    "AMPERSAND",
    "QUOTE",
    "LEFTPAREN",
    "RIGHTPAREN",
    "ASTERISK",
    "PLUS",
    "COMMA",
    "MINUS",
    "PERIOD",
    "SLASH",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "COLON",
    "SEMICOLON",
    "LESS",
    "EQUALS",
    "GREATER",
    "QUESTION",
    "AT",

    "LEFTBRACKET",
    "BACKSLASH",
    "RIGHTBRACKET",
    "CARET",
    "UNDERSCORE",
    "BACKQUOTE",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "DELETE",

    /* Keypad */
    "KP0",
    "KP1",
    "KP2",
    "KP3",
    "KP4",
    "KP5",
    "KP6",
    "KP7",
    "KP8",
    "KP9",
    "KP_PERIOD",
    "KP_DIVIDE",
    "KP_MULTIPLY",
    "KP_MINUS",
    "KP_PLUS",
    "KP_ENTER",
    "KP_EQUALS",

    /* Arrows + Home/End pad */
    "UP",
    "DOWN",
    "RIGHT",
    "LEFT",
    "INSERT",
    "HOME",
    "END",
    "PAGEUP",
    "PAGEDOWN",

    /* Function keys */
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "F11",
    "F12",
    "F13",
    "F14",
    "F15",

    /* Key state modifier keys */
    "NUMLOCK",
    "CAPSLOCK",
    "SCROLLOCK",
    "RSHIFT",
    "LSHIFT",
    "RCTRL",
    "LCTRL",
    "RALT",
    "LALT",
    "RMETA",
    "LMETA",
    "LSUPER",
    "RSUPER",
    "MODE",
    "COMPOSE",

    /* Miscellaneous function keys */
    "HELP",
    "PRINT",
    "SYSRQ",
    "BREAK",
    "MENU",
    NULL
  };

  //----------------------------------------------------------------------------
  ///
  /// Null-terminated array of mouse button names
  ///

  const char* button_names_mouse[] =
  {
    "UNKNOWN",
    NULL
  };

  //----------------------------------------------------------------------------
  ///
  /// Null-terminated array of joystick button names
  ///

  const char* button_names_joystick[] =
  {
    "UNKNOWN",
    NULL
  };

  //----------------------------------------------------------------------------
  // Implementation of class Device
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  Device::Device( Types type, const core::string& name, size_t num_buttons ):
    _name( name ),
    _num_buttons( num_buttons ),
    _buttons( new PhysicalButtonPtr[_num_buttons] )
  {
    const char** name_array = getButtonNames(type);

    // save wheather we reached the end of the name array (also check for NULL)
    bool array_end = (name_array == NULL);

    for( size_t i = 0; i < _num_buttons; ++i )
    {
      if( !array_end )
      {
        if( !name_array[i] )
          array_end = true;
        else
          // we have an name -> lets set it
          _buttons[i].reset( new PhysicalButton(name_array[i]) );
      }
      else
      {
        // no name found -> set to unknown
        _buttons[i].reset( new PhysicalButton("unknown") );
      }
    }
  }

  //----------------------------------------------------------------------------
  Device::~Device()
  {

  }

  //----------------------------------------------------------------------------
  PhysicalButtonPtr Device::getButton( size_t index )
  {
    assert( index < _num_buttons );

    return _buttons[index];
  }

  //----------------------------------------------------------------------------
  size_t Device::getNumButtons( Device::Types type )
  {
    const char** name_array = getButtonNames(type);

    if( !name_array )
      return 0;

    for( size_t count = 0;; ++count )
      if( name_array[count] == NULL )
        return count;
  }

  //----------------------------------------------------------------------------
  size_t Device::getId( Device::Types type, const core::string& name )
  {
    const char** name_array = getButtonNames(type);

    if( !name_array )
      return 0; // 0 == unknown

    for( size_t i = 0; name_array[i] != NULL; ++i )
      if( name == name_array[i] ) return i;

    LOG_FMT_WARNING_1("Button name not found (%1%)", name);

    return 0;
  }

  //----------------------------------------------------------------------------
  const char** Device::getButtonNames( Types type )
  {
    switch( type )
    {
      case KEYBOARD:
        return button_names_keyboard;
      case MOUSE:
        return button_names_mouse;
      case JOYSTICK:
        return button_names_joystick;
      default:
        return 0;
    }
  }

} // namespace input
} // namespace gewoopi
