//------------------------------------------------------------------------------
// manager.cpp - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/**
 * @file input/manager.cpp
 * @brief Input manager to support all kinds of input devices.
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 05.07.2009
 */

#include <iostream>
#include <stdexcept>

#include <iterator>

#include <xml-sipa/document.h>
#include <xml-sipa/element.h>
#include <xml-sipa/iterator.h>

#include "core/vfs/ifstream.hpp"

#include "input/manager.h"
#include "input/device.h"
#include "input/button.h"
#include "input/physical_button.hpp"

namespace gewoopi
{
namespace input
{

  //----------------------------------------------------------------------------
  Manager::Manager()
  {

  }

  //----------------------------------------------------------------------------
  Manager::~Manager()
  {

  }

  //----------------------------------------------------------------------------
  DeviceId Manager::createDevice( Device::Types type,
                                  const core::string& name,
                                  size_t num_buttons )
  {
    DevicePtr device( new input::Device(type, name, num_buttons) );

    _devices.push_back(device);

    // device id is index in the vector
    return _devices.size() - 1;
  }

  //----------------------------------------------------------------------------
  DevicePtr Manager::getDevice( DeviceId device )
  {
    assert( device < _devices.size() );

    return _devices.at(device);
  }

  //----------------------------------------------------------------------------
  DevicePtr Manager::getKeyboard()
  {
    assert( _devices.size() > 0 );

    return _devices.at(0); // 0 is always the system keyboard
  }

  //----------------------------------------------------------------------------
  DevicePtr Manager::getMouse()
  {
    assert( _devices.size() > 1 );

    return _devices.at(1); // 1 is always the system mouse
  }

  //----------------------------------------------------------------------------
  void Manager::notifyAbsoluteChange( DeviceId device,
                                      size_t button, float value )
  {
    assert( device < _devices.size() );

    // check for unknown button
    if( button < 0 ) return;

    _devices.at(device)->getButton(button)->setValue( value );
  }

  //----------------------------------------------------------------------------
  void Manager::notifyRelateiveChange( DeviceId device,
                                       size_t button, float value )
  {
    assert( device < _devices.size() );

    // check for unknown button
    if( button < 0 ) return

    _devices.at(device)->getButton(button)->addValue(value);
  }

  //----------------------------------------------------------------------------
  void Manager::addButton( const Button& button )
  {
    if( _virtual_buttons.find(button.getName()) == _virtual_buttons.end() )
    {
      _virtual_buttons[button.getName()] = button;
    }
    else
    {
      // TODO create assert log macro (create macro with tests)
      LOG_STR_WARNING_1("Duplicate Button (" + button.getName() + ")");
    }
  }

  //----------------------------------------------------------------------------
  void Manager::initVirtualButtons()
  {
    const std::string KEY_MAP_FILE("data/keymap.xml");
    core::vfs::ifstream key_map(KEY_MAP_FILE);

    xml::Document config;
    config.read( key_map );

    xml::ElementIterator root( config.root() );

    if( !root || root->getName() != "gewoopi-config" )
      throw std::runtime_error("No gewoopi config file.");
    if( root->getAttribute<std::string>("type") != "keymap" )
      throw std::runtime_error("No virtual keymap config file.");

    xml::ElementIterator vkey( root.first_child() );

    while( vkey != config.end() )
    {
      if( vkey->getName() == "vkey" )
      {
        // create a new virtual button
        Button virtual_button( vkey->getAttribute<core::string>("name") );

        // get all hardware buttons
        xml::ElementIterator hkey( vkey.first_child() );

        while( hkey != config.end() )
        {
          if( hkey->getName() == "hkey" )
          {
            const std::string& device_type = hkey->getAttribute("device");

            if( device_type == "keyboard" )
              virtual_button.add
              (
                getKeyboard()->getButton
                (
                  Device::getId( Device::KEYBOARD,
                                 hkey->getAttribute("button") )
                )
              );
          }

          hkey = hkey.next_sibling();
        }

        // and now add the virtual button
        addButton(virtual_button);

        LOG_FMT_INFO_2( "Createt virtual button (%1%)",
                        virtual_button.getName() );
      }

      vkey = vkey.next_sibling();
    }

    std::cout << "Successfully loaded " << KEY_MAP_FILE << std::endl;
  }

  //----------------------------------------------------------------------------
  const Button Manager::getButton( const core::string& name ) const
  {
    ButtonMap::const_iterator button( _virtual_buttons.find(name) );

    if( button != _virtual_buttons.end() )
      return button->second;
    else
      return Button();
  }


} // namespace input
} // namespace gewoopi
