//------------------------------------------------------------------------------
// plane.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 20.10.2009
 */

#include "math/plane.hpp"
#include "math/constants.hpp"

namespace gewoopi
{
namespace math
{

  //----------------------------------------------------------------------------
  Plane::Plane( const Vector3f& p0, const Vector3f& p1, const Vector3f& p2 ):
    _normal( Vector3f(p2 - p0).cross(p1 - p0).normalize() ),
    _d( p0 * _normal )
  {

  }

  //----------------------------------------------------------------------------
  Plane::Plane( const Vector3f& point, const Vector3f& normal ):
    _normal( normal.normalize() ),
    _d( point * _normal )
  {

  }

  //----------------------------------------------------------------------------
  Plane::Plane()
  {

  }

  //----------------------------------------------------------------------------
  const Vector3f& Plane::getNormal() const
  {
    return _normal;
  }

  //----------------------------------------------------------------------------
  real Plane::getD() const
  {
    return _d;
  }

  //----------------------------------------------------------------------------
  real Plane::distanceTo( const Vector3f& vec ) const
  {
    return vec * _normal - _d;
  }

  //----------------------------------------------------------------------------
  bool Plane::contains( const Vector3f& point ) const
  {
    return std::fabs( distanceTo(point) ) < epsilon<real>();
  }

  //----------------------------------------------------------------------------
  bool Plane::isBehind( const Vector3f& point ) const
  {
    return distanceTo(point) < -epsilon<real>();
  }

  //----------------------------------------------------------------------------
  bool Plane::hit( const Ray3f& ray,
                   real* distance,
                   Vector3f* normal ) const
  {
    real origin_dist = distanceTo(ray._origin);

    // check for the position of the origin
    if( origin_dist < epsilon<real>() )
    {
      if( origin_dist > -epsilon<real>() )
        // The point has a distance of less the epsilon to the plane,
        // therefore it is on the plane.
      {
        // set distance and normal if needed
        if( distance )
          *distance = 0;
        if( normal )
          *normal = _normal;

        return true;
      }
      else
      {
        // TODO maybe add the possibility to ignore this?
        // The origin is behind the plane.
        return false;
      }
    }

    //          normal * origin - d
    // dist = - -------------------
    //          normal * direction

    real nor_mul_dir = _normal * ray._direction;
    if( std::fabs(nor_mul_dir) < epsilon<real>() )
      // line is parallel to the plane and the origin is not on the plane
      // therefore no collision is possible.
      return false;

    // set distance and normal if needed
    if( distance )
      *distance = -origin_dist/nor_mul_dir;
    if( normal )
      *normal = _normal;

    return true;
  }

  //----------------------------------------------------------------------------
  std::ostream& operator <<( std::ostream& strm, const Plane& plane )
  {
    strm << "[normal=" << plane.getNormal() << ", d=" << plane.getD() << "]";
    return strm;
  }

} // namespace math
} // namespace gewoopi
