//------------------------------------------------------------------------------
// mesh.cpp - this file is part of the gewoopi game engine
// (c) copyright 2008 - 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#include <cstring>

#include "scene/mesh.hpp"
#include "video/renderer.h"

namespace gewoopi
{
namespace scene
{

  //----------------------------------------------------------------------------
  Mesh::Mesh( VertexArray vertices,
              Count vertex_count,
              const video::PrimitiveType primitive_type ):
    _vertices      ( vertices ),
    _vertex_count  ( vertex_count ),
    _primitive_type( primitive_type ),
    _face_count    ( 0 )
  {

  }

  //----------------------------------------------------------------------------
  Mesh::~Mesh()
  {

  }

  //----------------------------------------------------------------------------
  void Mesh::calculateCollisionData()
  {
    size_t offset = 0;

    switch( _primitive_type )
    {
      case video::TRIANGLE_LIST:
        offset = 3;
        break;
      // TODO handle other tyes as well
      default:
        //LOG_WARNING("Primitive type not supported for collsion detection");
        std::cout
          << "[WARNING] Primitive type not supported for collsion detection"
          << std::endl;
        break;
    }

    if( !offset )
      return;

    _face_count = _vertex_count / offset;

    if( !_face_count )
      return;

    _faces.reset( new math::Facef[_face_count] );

    for( size_t i = 0; i < _face_count; ++i )
    {
      _faces[i] = Face( _vertices[ i * 3     ].getPosition(),
                        _vertices[ i * 3 + 1 ].getPosition(),
                        _vertices[ i * 3 + 2 ].getPosition() );
    }
  }

  //----------------------------------------------------------------------------
  bool Mesh::hit( const math::Ray3f& ray,
                  math::real* distance,
                  math::Vector3f* normal ) const
  {
    math::real dist, min_dist = -1.f;

    for( size_t i = 0; i < _face_count; ++i )
    {
      if(    _faces[i].hit( ray, &dist, normal )
             // ignore hits from behind
          && dist > -math::epsilon<math::real>() )
      {
        if(    min_dist < -math::epsilon<math::real>()
            || dist < min_dist )
        {
          min_dist = dist;
        }
      }
    }

    if( min_dist > -1.f )
    {
      *distance = min_dist;
      return true;
    }

    return false;
  }

  //----------------------------------------------------------------------------
  void Mesh::render( video::RendererPtr renderer, int pass )
  {
    renderer->drawPrimitives( _primitive_type, _vertices, _vertex_count );
  }

  //----------------------------------------------------------------------------
  MeshPtr Mesh::createCube( float size )
  {
    float half_size = size / 2;
    const Count vert_count = 6*6;

    Vertex vertices[vert_count] =
    { // back
      math::Vector3f(  half_size, -half_size, -half_size ),
      math::Vector3f(  half_size,  half_size, -half_size ),
      math::Vector3f( -half_size,  half_size, -half_size ),

      math::Vector3f(  half_size, -half_size, -half_size ),
      math::Vector3f( -half_size,  half_size, -half_size ),
      math::Vector3f( -half_size, -half_size, -half_size ),

      // bottom
      math::Vector3f(  half_size, -half_size,  half_size ),
      math::Vector3f(  half_size, -half_size, -half_size ),
      math::Vector3f( -half_size, -half_size, -half_size ),

      math::Vector3f(  half_size, -half_size,  half_size ),
      math::Vector3f( -half_size, -half_size, -half_size ),
      math::Vector3f( -half_size, -half_size,  half_size ),

      // left
      math::Vector3f( -half_size,  half_size, -half_size ),
      math::Vector3f( -half_size,  half_size,  half_size ),
      math::Vector3f( -half_size, -half_size,  half_size ),

      math::Vector3f( -half_size,  half_size, -half_size ),
      math::Vector3f( -half_size, -half_size,  half_size ),
      math::Vector3f( -half_size, -half_size, -half_size ),

      // right
      math::Vector3f(  half_size,  half_size,  half_size ),
      math::Vector3f(  half_size,  half_size, -half_size ),
      math::Vector3f(  half_size, -half_size, -half_size ),

      math::Vector3f(  half_size,  half_size,  half_size ),
      math::Vector3f(  half_size, -half_size, -half_size ),
      math::Vector3f(  half_size, -half_size,  half_size ),

      // top
      math::Vector3f(  half_size,  half_size, -half_size ),
      math::Vector3f(  half_size,  half_size,  half_size ),
      math::Vector3f( -half_size,  half_size,  half_size ),

      math::Vector3f(  half_size,  half_size, -half_size ),
      math::Vector3f( -half_size,  half_size,  half_size ),
      math::Vector3f( -half_size,  half_size, -half_size ),

      // front
      math::Vector3f(  half_size,  half_size,  half_size ),
      math::Vector3f(  half_size, -half_size,  half_size ),
      math::Vector3f( -half_size, -half_size,  half_size ),

      math::Vector3f(  half_size,  half_size,  half_size ),
      math::Vector3f( -half_size, -half_size,  half_size ),
      math::Vector3f( -half_size,  half_size,  half_size )
    };

    VertexArray vert_array( new Vertex[vert_count] );

    for( Count i = 0; i < vert_count; ++i )
      vert_array[i] = vertices[i];

    return MeshPtr( new Mesh( vert_array, vert_count, video::TRIANGLE_LIST ) );
  }

} // namespace scene
} // namespace gewoopi
