//------------------------------------------------------------------------------
// loader_obj.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

/*!
 * @file
 * @brief
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 22.09.2009
 */

#define __SCOPE__ "scene"

#include "loader_obj.hpp"

#include "../../std.hpp"
#include "utilities/utilities.h"

#include <algorithm>
#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/spirit/include/classic_core.hpp>

namespace gewoopi
{
namespace scene
{

  //----------------------------------------------------------------------------
  LoaderOBJ::LoaderOBJ()
  {

  }

  //----------------------------------------------------------------------------
  LoaderOBJ::~LoaderOBJ()
  {

  }

  //----------------------------------------------------------------------------
  MeshPtr LoaderOBJ::load( const core::path& file )
  {
    LOG_STR_EVENT_3("Loading model: " + file.string() );

    iterator_t file_it( file.string() );

    if( !file_it )
      throw std::runtime_error("Failed to open " + file.string() );

    // reset parse variables
    _index_position = 0;
    _index_normal   = 0;
    _index_position_index = 0;
    _index_normal_index   = 0;

    _positions.clear();
    _normals.clear();
    _vertices.clear();

    using namespace boost::spirit::classic;
    using boost::bind;

    typedef scanner<iterator_t>     scanner_t;
    typedef rule<scanner_t>         rule_t;

    // create the parsing rules
    static rule_t const spaces
      =  *(blank_p - eol_p)
      ;
    static rule_t const vertex_position
      // v x y z [w]
      =  ch_p('v')
         >> +(  spaces
             >> real_p[ bind(&LoaderOBJ::addPositionCoord, this, _1) ]
             )
      ;
    static rule_t const vertex_normal
      // vn x y z [w]
      =  str_p("vn")
         >> +(  spaces
             >> real_p[ bind(&LoaderOBJ::addNormalCoord, this, _1) ]
             )
      ;
    static rule_t const face
      // f v/vt/vn v/vt/vn v/vt/vn [v/vt/vn]
      =  ch_p('f')
         >> +(  spaces >> int_p[ bind(&LoaderOBJ::addPositionIndex, this, _1) ]
             // texture coord and normal are optional
             >> !(  '/'
                 >> !int_p // ignore texture coord for the moment
                 >> '/'
                 >> !int_p[ bind(&LoaderOBJ::addNormalIndex, this, _1) ]
                 )
             )
      ;
    static rule_t const file_name
      = spaces >> +print_p
      ;
    static rule_t const mtllib
      // mtllib name.mtl
      = "mtllib"
         >> +file_name[ bind(&LoaderOBJ::addMatLib, this, _1, _2) ]
      ;
    static rule_t const comment
      // # comment...
      =  ch_p('#') >> *(anychar_p - eol_p)
      ;
    static rule_t const unknow
      =  +(anychar_p - eol_p)
      ;
    static rule_t const line
      =  (  vertex_position[ bind(&LoaderOBJ::addPosition, this, _1, _2) ]
         |  vertex_normal  [ bind(&LoaderOBJ::addNormal,   this, _1, _2) ]
         |  face           [ bind(&LoaderOBJ::addFace,     this, _1, _2) ]
         |  mtllib
         |  comment
         |  unknow
         )
         >> spaces >> *eol_p
         ;

    parse_info<iterator_t> info
      = parse( file_it, file_it.make_end(), *line );

    if( !info.full )
      throw std::runtime_error(
        "Failed to load modell (" + file.string() + ")" );

    // copy vertices to an VertexArrayPtr for the Mesh class
    VertexArray vert_array( new Vertex[_vertices.size()] );
    std::copy(_vertices.begin(),_vertices.end(),vert_array.get());

    LOG_STR_EVENT_1("Successfully loaded model: " + file.string() );

    return MeshPtr( new Mesh( vert_array, _vertices.size(),
                              video::TRIANGLE_LIST ) );
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addPositionCoord( float value )
  {
    if( _index_position < 3 )
      _tmp_position[_index_position] = value;

    ++_index_position;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addPosition( iterator_t, iterator_t const& )
  {
    if( _index_position < 3 )
      throw std::runtime_error( "Invalid Vector (num_values="
                                + utilities::toStr(_index_position) + ")" );

    // push back new vertex and reset current index
    _positions.push_back( _tmp_position );
    _index_position = 0;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addNormalCoord( float value )
  {
    if( _index_normal < 3 )
      _tmp_normal[_index_normal] = value;

    ++_index_normal;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addNormal( iterator_t, iterator_t const& )
  {
    if( _index_normal < 3 )
      throw std::runtime_error( "Invalid Normal (num_values="
                                + utilities::toStr(_index_normal) + ")" );

    // push back new vertex and reset current index
    _normals.push_back( _tmp_normal );
    _index_normal = 0;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addPositionIndex( int index )
  {
    assert( index > 0 ); // TODO maybe implement negative index

    if( _index_position_index < 4 )
    {
      // .obj start indices with 1...
      if( index > 0 )
        _tmp_position_index[_index_position_index] =
          static_cast<math::uint32>(index - 1);
    }

    ++_index_position_index;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addNormalIndex( int index )
  {
    assert( index > 0 ); // TODO maybe implement negative index

    if( _index_normal_index < 4 )
    {
      // .obj start indices with 1...
      if( index > 0 )
        _tmp_normal_index[_index_normal_index] =
          static_cast<math::uint32>(index - 1);
    }

    ++_index_normal_index;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addFace( iterator_t, iterator_t const& )
  {
    if( _index_position_index < 3 )
      throw std::runtime_error(
        "Invalid Face (num_positions="
        + utilities::toStr(_index_position_index) + ")" );

    if(    _index_normal_index
        && (_index_normal_index != _index_position_index) )
      throw std::runtime_error(
        "Invalid Face (num_normals="
        + utilities::toStr(_index_normal_index) + ")" );

    math::Vector3f normal;
    if( !_index_normal_index )
      // calculate a normal if non was read
      // calculation is reversed, because faces are stored ccw!
      normal =
      math::Vector3f
      (_positions[_tmp_position_index[1]] - _positions[_tmp_position_index[0]])
      .cross
      (_positions[_tmp_position_index[2]] - _positions[_tmp_position_index[0]])
      .normalize();


    // because the faces are stored ccw we need to go the
    // other way round
    if( _index_position_index == 3 )
    // we have a triangle
    {
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[2]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [2]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[1]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [1]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[0]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [0]]
                                         : normal) );
    }
    else
    // we have a quad (TODO are more then 4 indices possible?)
    // let's triangulate it...
    {
      // first triangle
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[2]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [2]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[1]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [1]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[0]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [0]]
                                         : normal) );

      // second triangle
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[3]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [3]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[2]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [2]]
                                         : normal) );
      _vertices.push_back( video::Vertex(_positions[_tmp_position_index[0]],
                                         _index_normal_index
                                         ? _normals[_tmp_normal_index  [0]]
                                         : normal) );
    }

    _index_position_index = 0;
    _index_normal_index   = 0;
  }

  //----------------------------------------------------------------------------
  void LoaderOBJ::addMatLib( iterator_t first, iterator_t const& last )
  {
    std::string name( first, last );
    std::cout << name << std::endl;
  }

} // namespace scene
} // namespace gewoopi
