//------------------------------------------------------------------------------
// node.cpp - this file is part of the gewoopi game engine
// (c) copyright 2009 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

#include "scene/node.h"
#include "video/renderer.h"

namespace gewoopi
{
namespace scene
{

  //----------------------------------------------------------------------------
  Node::Node( const math::Vector3f& position,
              const math::Vector3f& rotation ):
    _position_rel ( position ),
    _rotation_rel ( rotation ),
    _rel_changed ( true ),
    _abs_changed ( true )
  {

  }

  //----------------------------------------------------------------------------
  Node::~Node()
  {

  }

  //----------------------------------------------------------------------------
  void Node::render( video::RendererPtr renderer, int pass )
  {
    renderer->pushMatrix();
    renderer->multMatrix( _transform_abs );


    foreach( MeshPtr mesh, _meshes )
    {
      mesh->render( renderer, pass );
    }

    renderer->popMatrix();
  }

  //----------------------------------------------------------------------------
  void Node::move( double run_time,
                   double frame_time,
                   const math::Matrix4x4f& parent_transform,
                   bool parent_changed )
  {
    if( _rel_changed )
    {
      _transform_rel  =
        math::Matrix4x4f::RotateZ( math::Degreef(_rotation_rel.z()) );
      _transform_rel *=
        math::Matrix4x4f::RotateX( math::Degreef(_rotation_rel.x()) );
      _transform_rel *=
        math::Matrix4x4f::RotateY( math::Degreef(_rotation_rel.y()) );
      _transform_rel *=
        math::Matrix4x4f::Translate( _position_rel );
    }

    if( _rel_changed || parent_changed )
    {
      _transform_abs = parent_transform * _transform_rel;
      _abs_changed   = true;
    }

    _rel_changed = false;
  }

  //----------------------------------------------------------------------------
  bool Node::hasChanged()
  {
    return _abs_changed;
  }

  //----------------------------------------------------------------------------
  MeshPtr Node::addMesh( MeshPtr mesh )
  {
    if( mesh )
      _meshes.insert( mesh );
    return mesh;
  }

  //----------------------------------------------------------------------------
  bool Node::hit( const math::Ray3f& ray,
                  math::real* distance,
                  math::Vector3f* normal ) const
  {
    const math::Matrix4x4f inv_trans( _transform_abs.inverse() );
    const math::Vector3f orig = inv_trans * ray._origin;
    const math::Vector3f dir  = inv_trans * (ray._origin + ray._direction) - orig;

    foreach( MeshPtr mesh, _meshes )
    {
      // TODO transform length also back...
      if( mesh->hit(math::Ray3f(orig,dir), distance, normal) )
      {
        return true;
      }
    }

    return false;
  }

  //----------------------------------------------------------------------------
  void Node::setPosition( const math::Vector3f& position )
  {
    _position_rel = position;
    _rel_changed = true;
  }

  //----------------------------------------------------------------------------
  void Node::setRotation( const math::Vector3f& rotation )
  {
    _rotation_rel = rotation;
    _rel_changed = true;
  }

  //----------------------------------------------------------------------------
  void Node::addPosition( const math::Vector3f& move )
  {
    _position_rel += move;
    _rel_changed = true;
  }

  //----------------------------------------------------------------------------
  void Node::addRotation( const math::Vector3f& rotate )
  {
    _rotation_rel += rotate;
    _rel_changed = true;
  }

  //----------------------------------------------------------------------------
  const math::Matrix4x4f& Node::getTransformAbs() const
  {
    return _transform_abs;
  }

} // namespace scene
} // namespace gewoopi
