//------------------------------------------------------------------------------
// OpenGL/renderer.cpp - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include "renderer.h"
#include "math/matrix.hpp"
#include "scene/camera.h"

namespace gewoopi
{
namespace video
{

  //----------------------------------------------------------------------------
  OGLRenderer::OGLRenderer()
  {
    LOG_EVENT("Create OpenGL Renderer.");
  }

  OGLRenderer::~OGLRenderer()
  {

  }

  //----------------------------------------------------------------------------
  math::Color3f OGLRenderer::setClearColor( const math::Color3f& color )
  {
    glClearColor( color.r(), color.g(), color.b(), 1 );

    return Renderer::setClearColor( color );
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::loadMatrix( const math::Matrix4x4f& mat )
  {
    glLoadMatrixf( mat );
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::multMatrix( const math::Matrix4x4f& mat )
  {
    glMultMatrixf( mat );
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::pushMatrix()
  {
    glPushMatrix();
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::popMatrix()
  {
    glPopMatrix();
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::beginScene( bool clear_back_buffer, bool clear_z_buffer )
  {
    GLbitfield mask =   (clear_back_buffer ? GL_COLOR_BUFFER_BIT : 0)
                      | (clear_z_buffer    ? GL_DEPTH_BUFFER_BIT : 0);

    glClear( mask );

    // setup projection
    glMatrixMode( GL_PROJECTION );
    glLoadMatrixf( _camera->getProjection() );

    // setup transform
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    float lp[]={0,2,-3,1};
    float ld[]={0.5,-0.5,-1,1};

    glLightfv(GL_LIGHT0,GL_POSITION,lp);
    glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,ld);
    glLoadMatrixf( _camera->getTransform() );
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::endScene()
  {
    SDL_GL_SwapBuffers();

    GLenum error;

    while( (error = glGetError()) != GL_NO_ERROR )
    {
      switch( error )
      {
        case GL_INVALID_ENUM:
          std::cerr << "GL_INVALID_ENUM" << std::endl;
          break;
        case GL_INVALID_VALUE:
          std::cerr << "GL_INVALID_VALUE" << std::endl;
          break;
        case GL_INVALID_OPERATION:
          std::cerr << "GL_INVALID_OPERATION" << std::endl;
          break;
        case GL_STACK_OVERFLOW:
          std::cerr << "GL_STACK_OVERFLOW" << std::endl;
          break;
        case GL_STACK_UNDERFLOW:
          std::cerr << "GL_STACK_UNDERFLOW" << std::endl;
          break;
        case GL_OUT_OF_MEMORY:
          std::cerr << "GL_OUT_OF_MEMORY" << std::endl;
          break;
      }
    }
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::drawPrimitives(  const PrimitiveType type,
                                     scene::VertexArray vertices,
                                     scene::Count count )
  {
    scene::Vertex* vert = vertices.get();
    scene::Count counter = count;

    glBegin( getPrimitiveType(type) );
      while( counter-- )
      {
        glNormal3fv( vert->getNormal()   );
        glColor3fv ( vert->getColor()    );
        glVertex3fv( vert->getPosition() );
        ++vert;
      }
    glEnd();

    // render normals...
    /*vert = vertices.get();

    glColor3fv( math::Color3f::RED );

    glBegin( GL_LINES );
      while( count-- )
      {
        glVertex3fv( vert->getPosition() );
        glVertex3fv( vert->getPosition() + vert->getNormal() );
        ++vert;
      }
    glEnd();*/
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::drawIndexedPrimitives( const PrimitiveType type,
                                           scene::VertexArray vertices,
                                           scene::Count count,
                                           const math::Color3f *color )
  {

  }

  //----------------------------------------------------------------------------
  void OGLRenderer::performOperation( RenderOperation& operation )
  {

  }

  //----------------------------------------------------------------------------
  void OGLRenderer::setCullMode( const CullMode mode )
  {
    switch( mode )
    {
      case NONE:
        glDisable(GL_CULL_FACE);
        break;
      case CW:
        glEnable(GL_CULL_FACE);
        glFrontFace(GL_CCW);
        break;
      case CCW:
        glEnable(GL_CULL_FACE);
        glFrontFace(GL_CW);
        break;
    }
  }

  //----------------------------------------------------------------------------
  void OGLRenderer::setMaterial( const Material& mat )
  {
    glMaterialfv( GL_FRONT, GL_DIFFUSE,  mat.getDiffuse().getPointer() );
    glMaterialfv( GL_FRONT, GL_AMBIENT,  mat.getAmbient().getPointer() );
    glMaterialfv( GL_FRONT, GL_SPECULAR, mat.getSpecular().getPointer() );
    glMaterialfv( GL_FRONT, GL_EMISSION, mat.getEmissive().getPointer() );
  }

  //----------------------------------------------------------------------------
  unsigned int OGLRenderer::getPrimitiveType( const PrimitiveType type )
  {
    switch( type )
    {
      case TRIANGLE_LIST:
        return GL_TRIANGLES;
      case TRIANGLE_STRIP:
        return GL_TRIANGLE_STRIP;
      case TRIANGLE_FAN:
        return GL_TRIANGLE_FAN;
      case QUAD_LIST:
        return GL_QUADS;
      default:
        throw std::exception(); // TODO better exception handling
    }
  }

} // namespace video
} // namespace gewoopi
