//------------------------------------------------------------------------------
// OpenGL/window.cpp - this file is part of the gewoopi game engine
// (c) copyright 2008 Thomas Geymayer <tomgey@gmail.com>
// For conditions of distribution and use, see copyright notice in gewoopi.h
//------------------------------------------------------------------------------

#include <string>
#include <boost/format.hpp>

#define NO_SDL_GLEXT
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include "math/vector.hpp"
#include "input/manager.h"
#include "input/device.h"
#include "input/button.h"
#include "window.h"

namespace gewoopi
{
namespace video
{
  int keyboard_keymap[SDLK_LAST];

  //----------------------------------------------------------------------------
  OGLWindow::OGLWindow( const math::Size2D<int>& size,
                        SurfaceFormat format,
                        input::ManagerPtr input_manager ):
    Window( size, format ),
    _screen( NULL ),
    _input_manager( input_manager )
  {
    // TODO
    /// @todo Load from external file -> xml?

    using input::Device;
    Device::Types type = Device::KEYBOARD;

    keyboard_keymap[SDLK_UNKNOWN]    = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_BACKSPACE]  = Device::getId(type,"BACKSPACE");
    keyboard_keymap[SDLK_TAB]        = Device::getId(type,"TAB");
    keyboard_keymap[SDLK_CLEAR]      = Device::getId(type,"CLEAR");
    keyboard_keymap[SDLK_RETURN]     = Device::getId(type,"RETURN");
    keyboard_keymap[SDLK_PAUSE]      = Device::getId(type,"PAUSE");
    keyboard_keymap[SDLK_ESCAPE]     = Device::getId(type,"ESCAPE");
    keyboard_keymap[SDLK_SPACE]      = Device::getId(type,"SPACE");
    keyboard_keymap[SDLK_EXCLAIM]    = Device::getId(type,"EXCLAIM");
    keyboard_keymap[SDLK_QUOTEDBL]   = Device::getId(type,"QUOTEDBL");
    keyboard_keymap[SDLK_HASH]       = Device::getId(type,"HASH");
    keyboard_keymap[SDLK_DOLLAR]     = Device::getId(type,"DOLLAR");
    keyboard_keymap[SDLK_AMPERSAND]  = Device::getId(type,"AMPERSAND");
    keyboard_keymap[SDLK_QUOTE]      = Device::getId(type,"QUOTE");
    keyboard_keymap[SDLK_LEFTPAREN]  = Device::getId(type,"LEFTPAREN");
    keyboard_keymap[SDLK_RIGHTPAREN] = Device::getId(type,"RIGHTPAREN");
    keyboard_keymap[SDLK_ASTERISK]   = Device::getId(type,"ASTERISK");
    keyboard_keymap[SDLK_PLUS]       = Device::getId(type,"PLUS");
    keyboard_keymap[SDLK_COMMA]      = Device::getId(type,"COMMA");
    keyboard_keymap[SDLK_MINUS]      = Device::getId(type,"MINUS");
    keyboard_keymap[SDLK_PERIOD]     = Device::getId(type,"PERIOD");
    keyboard_keymap[SDLK_SLASH]      = Device::getId(type,"SLASH");
    keyboard_keymap[SDLK_0]          = Device::getId(type,"0");
    keyboard_keymap[SDLK_1]          = Device::getId(type,"1");
    keyboard_keymap[SDLK_2]          = Device::getId(type,"2");
    keyboard_keymap[SDLK_3]          = Device::getId(type,"3");
    keyboard_keymap[SDLK_4]          = Device::getId(type,"4");
    keyboard_keymap[SDLK_5]          = Device::getId(type,"5");
    keyboard_keymap[SDLK_6]          = Device::getId(type,"6");
    keyboard_keymap[SDLK_7]          = Device::getId(type,"7");
    keyboard_keymap[SDLK_8]          = Device::getId(type,"8");
    keyboard_keymap[SDLK_9]          = Device::getId(type,"9");
    keyboard_keymap[SDLK_COLON]      = Device::getId(type,"COLON");
    keyboard_keymap[SDLK_SEMICOLON]  = Device::getId(type,"SEMICOLON");
    keyboard_keymap[SDLK_LESS]       = Device::getId(type,"LESS");
    keyboard_keymap[SDLK_EQUALS]     = Device::getId(type,"EQUALS");
    keyboard_keymap[SDLK_GREATER]    = Device::getId(type,"GREATER");
    keyboard_keymap[SDLK_QUESTION]   = Device::getId(type,"QUESTION");
    keyboard_keymap[SDLK_AT]         = Device::getId(type,"AT");

    keyboard_keymap[SDLK_LEFTBRACKET]    = Device::getId(type,"LEFTBRACKET");
    keyboard_keymap[SDLK_BACKSLASH]      = Device::getId(type,"BACKSLASH");
    keyboard_keymap[SDLK_RIGHTBRACKET]   = Device::getId(type,"RIGHTBRACKET");
    keyboard_keymap[SDLK_CARET]          = Device::getId(type,"CARET");
    keyboard_keymap[SDLK_UNDERSCORE]     = Device::getId(type,"UNDERSCORE");
    keyboard_keymap[SDLK_BACKQUOTE]      = Device::getId(type,"BACKQUOTE");
    keyboard_keymap[SDLK_a]              = Device::getId(type,"A");
    keyboard_keymap[SDLK_b]              = Device::getId(type,"B");
    keyboard_keymap[SDLK_c]              = Device::getId(type,"C");
    keyboard_keymap[SDLK_d]              = Device::getId(type,"D");
    keyboard_keymap[SDLK_e]              = Device::getId(type,"E");
    keyboard_keymap[SDLK_f]              = Device::getId(type,"F");
    keyboard_keymap[SDLK_g]              = Device::getId(type,"G");
    keyboard_keymap[SDLK_h]              = Device::getId(type,"H");
    keyboard_keymap[SDLK_i]              = Device::getId(type,"I");
    keyboard_keymap[SDLK_j]              = Device::getId(type,"J");
    keyboard_keymap[SDLK_k]              = Device::getId(type,"K");
    keyboard_keymap[SDLK_l]              = Device::getId(type,"L");
    keyboard_keymap[SDLK_m]              = Device::getId(type,"M");
    keyboard_keymap[SDLK_n]              = Device::getId(type,"N");
    keyboard_keymap[SDLK_o]              = Device::getId(type,"O");
    keyboard_keymap[SDLK_p]              = Device::getId(type,"P");
    keyboard_keymap[SDLK_q]              = Device::getId(type,"Q");
    keyboard_keymap[SDLK_r]              = Device::getId(type,"R");
    keyboard_keymap[SDLK_s]              = Device::getId(type,"S");
    keyboard_keymap[SDLK_t]              = Device::getId(type,"T");
    keyboard_keymap[SDLK_u]              = Device::getId(type,"U");
    keyboard_keymap[SDLK_v]              = Device::getId(type,"V");
    keyboard_keymap[SDLK_w]              = Device::getId(type,"W");
    keyboard_keymap[SDLK_x]              = Device::getId(type,"X");
    keyboard_keymap[SDLK_y]              = Device::getId(type,"Y");
    keyboard_keymap[SDLK_z]              = Device::getId(type,"Z");
    keyboard_keymap[SDLK_DELETE]         = Device::getId(type,"DELETE");

    keyboard_keymap[SDLK_WORLD_0]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_1]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_2]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_3]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_4]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_5]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_6]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_7]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_8]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_9]        = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_10]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_11]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_12]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_13]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_14]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_15]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_16]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_17]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_18]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_19]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_20]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_21]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_22]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_23]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_24]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_25]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_26]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_27]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_28]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_29]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_30]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_31]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_32]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_33]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_34]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_35]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_36]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_37]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_38]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_39]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_40]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_41]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_42]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_43]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_44]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_45]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_46]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_47]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_48]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_49]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_50]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_51]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_52]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_53]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_54]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_55]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_56]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_57]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_58]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_59]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_60]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_61]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_62]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_63]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_64]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_65]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_66]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_67]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_68]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_69]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_70]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_71]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_72]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_73]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_74]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_75]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_76]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_77]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_78]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_79]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_80]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_81]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_82]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_83]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_84]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_85]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_86]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_87]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_88]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_89]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_90]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_91]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_92]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_93]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_94]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_WORLD_95]       = Device::getId(type,"UNKNOWN");

    keyboard_keymap[SDLK_KP0]         = Device::getId(type,"KP0");
    keyboard_keymap[SDLK_KP1]         = Device::getId(type,"KP1");
    keyboard_keymap[SDLK_KP2]         = Device::getId(type,"KP2");
    keyboard_keymap[SDLK_KP3]         = Device::getId(type,"KP3");
    keyboard_keymap[SDLK_KP4]         = Device::getId(type,"KP4");
    keyboard_keymap[SDLK_KP5]         = Device::getId(type,"KP5");
    keyboard_keymap[SDLK_KP6]         = Device::getId(type,"KP6");
    keyboard_keymap[SDLK_KP7]         = Device::getId(type,"KP7");
    keyboard_keymap[SDLK_KP8]         = Device::getId(type,"KP8");
    keyboard_keymap[SDLK_KP9]         = Device::getId(type,"KP9");
    keyboard_keymap[SDLK_KP_PERIOD]   = Device::getId(type,"KP_PERIOD");
    keyboard_keymap[SDLK_KP_DIVIDE]   = Device::getId(type,"KP_DIVIDE");
    keyboard_keymap[SDLK_KP_MULTIPLY] = Device::getId(type,"KP_MULTIPLY");
    keyboard_keymap[SDLK_KP_MINUS]    = Device::getId(type,"KP_MINUS");
    keyboard_keymap[SDLK_KP_PLUS]     = Device::getId(type,"KP_PLUS");
    keyboard_keymap[SDLK_KP_ENTER]    = Device::getId(type,"KP_ENTER");
    keyboard_keymap[SDLK_KP_EQUALS]   = Device::getId(type,"KP_EQUALS");

    keyboard_keymap[SDLK_UP]         = Device::getId(type,"UP");
    keyboard_keymap[SDLK_DOWN]       = Device::getId(type,"DOWN");
    keyboard_keymap[SDLK_RIGHT]      = Device::getId(type,"RIGHT");
    keyboard_keymap[SDLK_LEFT]       = Device::getId(type,"LEFT");
    keyboard_keymap[SDLK_INSERT]     = Device::getId(type,"INSERT");
    keyboard_keymap[SDLK_HOME]       = Device::getId(type,"HOME");
    keyboard_keymap[SDLK_END]        = Device::getId(type,"END");
    keyboard_keymap[SDLK_PAGEUP]     = Device::getId(type,"PAGEUP");
    keyboard_keymap[SDLK_PAGEDOWN]   = Device::getId(type,"PAGEDOWN");

    keyboard_keymap[SDLK_F1]         = Device::getId(type,"F1");
    keyboard_keymap[SDLK_F2]         = Device::getId(type,"F2");
    keyboard_keymap[SDLK_F3]         = Device::getId(type,"F3");
    keyboard_keymap[SDLK_F4]         = Device::getId(type,"F4");
    keyboard_keymap[SDLK_F5]         = Device::getId(type,"F5");
    keyboard_keymap[SDLK_F6]         = Device::getId(type,"F6");
    keyboard_keymap[SDLK_F7]         = Device::getId(type,"F7");
    keyboard_keymap[SDLK_F8]         = Device::getId(type,"F8");
    keyboard_keymap[SDLK_F9]         = Device::getId(type,"F9");
    keyboard_keymap[SDLK_F10]        = Device::getId(type,"F10");
    keyboard_keymap[SDLK_F11]        = Device::getId(type,"F11");
    keyboard_keymap[SDLK_F12]        = Device::getId(type,"F12");
    keyboard_keymap[SDLK_F13]        = Device::getId(type,"F13");
    keyboard_keymap[SDLK_F14]        = Device::getId(type,"F14");
    keyboard_keymap[SDLK_F15]        = Device::getId(type,"F15");

    keyboard_keymap[SDLK_NUMLOCK]    = Device::getId(type,"NUMLOCK");
    keyboard_keymap[SDLK_CAPSLOCK]   = Device::getId(type,"CAPSLOCK");
    keyboard_keymap[SDLK_SCROLLOCK]  = Device::getId(type,"SCROLLOCK");
    keyboard_keymap[SDLK_RSHIFT]     = Device::getId(type,"RSHIFT");
    keyboard_keymap[SDLK_LSHIFT]     = Device::getId(type,"LSHIFT");
    keyboard_keymap[SDLK_RCTRL]      = Device::getId(type,"RCTRL");
    keyboard_keymap[SDLK_LCTRL]      = Device::getId(type,"LCTRL");
    keyboard_keymap[SDLK_RALT]       = Device::getId(type,"RALT");
    keyboard_keymap[SDLK_LALT]       = Device::getId(type,"LALT");
    keyboard_keymap[SDLK_RMETA]      = Device::getId(type,"RMETA");
    keyboard_keymap[SDLK_LMETA]      = Device::getId(type,"LMETA");
    keyboard_keymap[SDLK_LSUPER]     = Device::getId(type,"LSUPER");
    keyboard_keymap[SDLK_RSUPER]     = Device::getId(type,"RSUPER");
    keyboard_keymap[SDLK_MODE]       = Device::getId(type,"MODE");
    keyboard_keymap[SDLK_COMPOSE]    = Device::getId(type,"COMPOSE");

    keyboard_keymap[SDLK_HELP]       = Device::getId(type,"HELP");
    keyboard_keymap[SDLK_PRINT]      = Device::getId(type,"PRINT");
    keyboard_keymap[SDLK_SYSREQ]     = Device::getId(type,"SYSRQ");
    keyboard_keymap[SDLK_BREAK]      = Device::getId(type,"BREAK");
    keyboard_keymap[SDLK_MENU]       = Device::getId(type,"MENU");
    keyboard_keymap[SDLK_POWER]      = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_EURO]       = Device::getId(type,"UNKNOWN");
    keyboard_keymap[SDLK_UNDO]       = Device::getId(type,"UNKNOWN");
  }

  //----------------------------------------------------------------------------
  OGLWindow::~OGLWindow()
  {
    SDL_Quit();
  }

  //----------------------------------------------------------------------------
  bool OGLWindow::load()
  {
    if( !SDL_WasInit( SDL_INIT_VIDEO ) )
      SDL_InitSubSystem( SDL_INIT_VIDEO );

    setCaption( getCaption(), true );

    SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE,     8 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE,   8 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,    8 );

    SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 1 );
    SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, 4 );


    _screen = SDL_SetVideoMode( getSize().width(), getSize().height(), 0,
                                SDL_OPENGL /*| SDL_FULLSCREEN*/ );

//    _screen = SDL_SetVideoMode( 1680, 1050, 0,
//                                    SDL_OPENGL | SDL_FULLSCREEN );

    if( !_screen )
    {
      LOG_ERROR("Failed to set the video mode.");
      throw std::runtime_error("Failed to set video mode.");
    }

    LOG_FMT_INFO( "Created OpenGL %1% context.\nDeviceinfo: %2% %3%",
                    glGetString(GL_VERSION)
                  % glGetString(GL_VENDOR)
                  % glGetString(GL_RENDERER) );

    std::cout << "Created OpenGL " << glGetString(GL_VERSION)
              << " device." << std::endl;

    glEnable( GL_DEPTH_TEST );
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glEnable( GL_COLOR_MATERIAL );
    glEnable( GL_MULTISAMPLE_ARB );

    float amb[]={0.2,0.1,0.2};
    glLightfv(GL_LIGHT0,GL_AMBIENT,amb);
/*    glLightfv(GL_LIGHT0, GL_POSITION, pos.getPointer() );
    glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, math::Vector3f(0,0,10).getPointer() );

    glLightfv(GL_LIGHT1, GL_POSITION, math::Vector3f(0,0,-10).getPointer() );
    glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, math::Vector3f(1,-1,10).getPointer() );
*/

    //glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

    _alive = true;

    initInputDevices();

    return true;
  }

  //----------------------------------------------------------------------------
  bool OGLWindow::updateCaption( const std::string& caption )
  {
    SDL_WM_SetCaption( caption.c_str(), NULL );

    return true;
  }


  //----------------------------------------------------------------------------
  void OGLWindow::processMessages()
  {
    SDL_Event event;

    while( SDL_PollEvent(&event) )
    {
      switch( event.type )
      {
        case SDL_QUIT:
          _alive = false;
          break;
        case SDL_ACTIVEEVENT:
          if( event.active.state & SDL_APPACTIVE );
          break;
        case SDL_VIDEOEXPOSE:
          break;
        case SDL_KEYDOWN:
        case SDL_KEYUP:
            _input_manager
              // event.key.state is 0 if released and 1 if pressed
              ->notifyAbsoluteChange( _keyboard_id,
                                      keyboard_keymap[event.key.keysym.sym],
                                      event.key.state );
            /*event.key.keysym.sym,
            event.key.state,
            event.key.which;*/
          break;
        case SDL_MOUSEBUTTONDOWN:
        case SDL_MOUSEBUTTONUP:
            /*event.button.button,
            event.button.state,
            event.button.which;*/
          break;
        case SDL_MOUSEMOTION:
            /*event.motion.xrel,
            event.motion.yrel,
            event.motion.which;*/
          break;
        default:
          std::cout << "[OGL-Window] Warning: Unknown message type."
                    << std::endl;
      }
    }
  }

  //----------------------------------------------------------------------------
  void OGLWindow::initInputDevices()
  {
    // register system keyboard and mouse
    _keyboard_id =
      _input_manager->createDevice(
          input::Device::KEYBOARD,
          "System Keyboard",
          input::Device::getNumButtons(input::Device::KEYBOARD) );

    _mouse_id =
      _input_manager->createDevice(
        input::Device::MOUSE,
        "System Mouse",
        input::Device::getNumButtons(input::Device::MOUSE) );

    // TODO enumerate joysticks and gamepads and also add them
  }

} // namespace video
} // namespace gewoopi
